from __future__ import annotations

import math
from abc import abstractmethod, ABC
from enum import IntEnum
from typing import List, Optional

from pykotor.extract.file import ResourceIdentifier

from pykotor.common.geometry import Vector2, Vector3, Vector4, Polygon3
from pykotor.common.language import LocalizedString
from pykotor.common.misc import Game, Color, ResRef
from pykotor.resource.formats.gff import GFF, GFFStruct, GFFList, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.generics.utc import bytes_utc, UTC
from pykotor.resource.generics.utd import UTD, bytes_utd
from pykotor.resource.generics.ute import UTE, bytes_ute
from pykotor.resource.generics.utm import UTM, bytes_utm
from pykotor.resource.generics.utp import bytes_utp, UTP
from pykotor.resource.generics.uts import UTS, bytes_uts
from pykotor.resource.generics.utt import bytes_utt, UTT
from pykotor.resource.generics.utw import UTW, bytes_utw
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class GIT:
    BINARY_TYPE = ResourceType.GIT

    def __init__(
            self
    ):
        self.ambient_sound_id: int = 0
        self.ambient_volume: int = 0
        self.env_audio: int = 0
        self.music_standard_id: int = 0
        self.music_battle_id: int = 0
        self.music_delay: int = 0

        self.cameras: List[GITCamera] = []
        self.creatures: List[GITCreature] = []
        self.doors: List[GITDoor] = []
        self.encounters: List[GITEncounter] = []
        self.placeables: List[GITPlaceable] = []
        self.sounds: List[GITSound] = []
        self.stores: List[GITStore] = []
        self.triggers: List[GITTrigger] = []
        self.waypoints: List[GITWaypoint] = []

    def instances(
            self
    ) -> List[GITInstance]:
        """
        Returns a list of all instances stored inside the GIT, regardless of the type.

        Returns:
            A list of all stored instances.
        """
        instances = []
        # We could just add these all together rather than using the extend method, but then PyCharms would get cranky
        # about the type hints...
        instances.extend(self.cameras)
        instances.extend(self.creatures)
        instances.extend(self.doors)
        instances.extend(self.encounters)
        instances.extend(self.placeables)
        instances.extend(self.sounds)
        instances.extend(self.stores)
        instances.extend(self.triggers)
        instances.extend(self.waypoints)
        return instances

    def remove(
            self,
            instance: GITInstance
    ):
        if isinstance(instance, GITCreature):
            self.creatures.remove(instance)
        elif isinstance(instance, GITPlaceable):
            self.placeables.remove(instance)
        elif isinstance(instance, GITDoor):
            self.doors.remove(instance)
        elif isinstance(instance, GITTrigger):
            self.triggers.remove(instance)
        elif isinstance(instance, GITEncounter):
            self.encounters.remove(instance)
        elif isinstance(instance, GITWaypoint):
            self.waypoints.remove(instance)
        elif isinstance(instance, GITCamera):
            self.cameras.remove(instance)
        elif isinstance(instance, GITSound):
            self.sounds.remove(instance)
        elif isinstance(instance, GITStore):
            self.stores.remove(instance)

    def index(
            self,
            instance: GITInstance
    ) -> int:
        """
        Finds the index of an instance in the particular list it belongs to inside the GIT object.

        Args:
            instance: The instance to search for.

        Raises:
            ValueError: If the given instance does not belong to the GIT.

        Returns:
            The index into one of the GIT instance lists.
        """
        try:
            if isinstance(instance, GITCreature):
                return self.creatures.index(instance)
            elif isinstance(instance, GITPlaceable):
                return self.placeables.index(instance)
            elif isinstance(instance, GITDoor):
                return self.doors.index(instance)
            elif isinstance(instance, GITTrigger):
                return self.triggers.index(instance)
            elif isinstance(instance, GITEncounter):
                return self.encounters.index(instance)
            elif isinstance(instance, GITWaypoint):
                return self.waypoints.index(instance)
            elif isinstance(instance, GITCamera):
                return self.cameras.index(instance)
            elif isinstance(instance, GITSound):
                return self.sounds.index(instance)
            elif isinstance(instance, GITStore):
                return self.stores.index(instance)
            else:
                raise ValueError
        except ValueError:
            raise ValueError("Could not find instance in GIT object.")

    def add(
            self,
            instance: GITInstance
    ) -> None:
        """
        Adds instance to the relevant list in the GIT.

        Args:
            instance: The instance to add into the GIT.

        Raises:
            ValueError: If the instance already is stored inside the GIT.
        """
        if isinstance(instance, GITCreature):
            if instance in self.creatures:
                ValueError("Creature instance already exists inside the GIT object.")
            return self.creatures.append(instance)
        elif isinstance(instance, GITPlaceable):
            if instance in self.placeables:
                ValueError("Placeable instance already exists inside the GIT object.")
            return self.placeables.append(instance)
        elif isinstance(instance, GITDoor):
            if instance in self.doors:
                ValueError("Door instance already exists inside the GIT object.")
            return self.doors.append(instance)
        elif isinstance(instance, GITTrigger):
            if instance in self.triggers:
                ValueError("Trigger instance already exists inside the GIT object.")
            return self.triggers.append(instance)
        elif isinstance(instance, GITEncounter):
            if instance in self.encounters:
                ValueError("Encounter instance already exists inside the GIT object.")
            return self.encounters.append(instance)
        elif isinstance(instance, GITWaypoint):
            if instance in self.waypoints:
                ValueError("Waypoint instance already exists inside the GIT object.")
            return self.waypoints.append(instance)
        elif isinstance(instance, GITCamera):
            if instance in self.cameras:
                ValueError("Camera instance already exists inside the GIT object.")
            return self.cameras.append(instance)
        elif isinstance(instance, GITSound):
            if instance in self.sounds:
                ValueError("Sound instance already exists inside the GIT object.")
            return self.sounds.append(instance)
        elif isinstance(instance, GITStore):
            if instance in self.stores:
                ValueError("Store instance already exists inside the GIT object.")
            return self.stores.append(instance)
        else:
            TypeError("Tired to add invalid instance.")


class GITInstance(ABC):
    def __init__(self, x: float, y: float, z: float):
        self.position: Vector3 = Vector3(x, y, z)

    @abstractmethod
    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        ...

    @abstractmethod
    def blank(
            self
    ) -> Optional[bytes]:
        ...

    @abstractmethod
    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        ...

    @abstractmethod
    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        ...

    @abstractmethod
    def classification(
            self
    ) -> str:
        ...

    def yaw(
            self
    ) -> Optional[float]:
        """
        Returns the yaw rotation (in radians) of the instance if the instance supports it, otherwise returns None.
        """


class GITCamera(GITInstance):
    GFF_STRUCT_ID = 14

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0,
            yaw: float = 0.0,
            pitch: float = 0.0,
            roll: float = 0.0,
            camera_id: int = 0
    ):
        super().__init__(x, y, z)
        self.camera_id = camera_id
        self.fov: float = 45
        self.height: float = 0.0
        self.mic_range: float = 0.0
        self.pitch: float = 0.0
        self.orientation: Vector4 = Vector4.from_euler(math.pi/2-yaw, roll, math.pi-pitch)

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        rotation = self.orientation.to_euler()
        rotation.x += yaw
        rotation.y += roll
        rotation.z += pitch
        self.orientation = Vector4.from_euler(rotation.x, rotation.y, rotation.z)

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return None

    def blank(
            self
    ) -> Optional[bytes]:
        return None

    def classification(
            self
    ) -> str:
        return "Camera"

    def yaw(
            self
    ) -> Optional[float]:
        return math.pi-self.orientation.to_euler().x


class GITCreature(GITInstance):
    GFF_STRUCT_ID = 4

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        super().__init__(x, y, z)
        self.resref: ResRef = ResRef.from_blank()
        self.bearing: float = 0.0

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        self.bearing += yaw

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return ResourceIdentifier(self.resref.get(), ResourceType.UTC)

    def blank(
            self
    ) -> Optional[bytes]:
        return bytes_utc(UTC())

    def extension(
            self
    ) -> Optional[ResourceType]:
        return ResourceType.UTC

    def classification(
            self
    ) -> str:
        return "Creature"

    def yaw(
            self
    ) -> Optional[float]:
        return self.bearing


class GITModuleLink(IntEnum):
    NoLink = 0
    ToDoor = 1
    ToWaypoint = 2


class GITDoor(GITInstance):
    GFF_STRUCT_ID = 8

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        super().__init__(x, y, z)
        self.resref: ResRef = ResRef.from_blank()
        self.bearing: float = 0.0
        self.tweak_color: Optional[Color] = None
        self.linked_to: str = ""
        self.linked_to_flags: GITModuleLink = GITModuleLink.NoLink
        self.linked_to_module: ResRef = ResRef.from_blank()
        self.transition_destination: LocalizedString = LocalizedString.from_invalid()
        self.tag: str = ""

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        self.bearing += yaw

    def blank(
            self
    ) -> Optional[bytes]:
        return bytes_utd(UTD())

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return ResourceIdentifier(self.resref.get(), ResourceType.UTD)

    def classification(
            self
    ) -> str:
        return "Door"

    def yaw(
            self
    ) -> Optional[float]:
        return self.bearing


class GITEncounterSpawnPoint:
    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        self.x = x
        self.y = y
        self.z = z
        self.orientation: float = 0.0


class GITEncounter(GITInstance):
    GFF_STRUCT_ID = 7
    GFF_GEOMETRY_STRUCT_ID = 1
    GFF_SPAWN_STRUCT_ID = 2

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        super().__init__(x, y, z)
        self.geometry: Polygon3 = Polygon3()
        self.spawn_points: List[GITEncounterSpawnPoint] = []
        self.resref: ResRef = ResRef.from_blank()

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        raise ValueError("Encounters cannot be rotated.")

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return ResourceIdentifier(self.resref.get(), ResourceType.UTE)

    def blank(
            self
    ) -> Optional[bytes]:
        return bytes_ute(UTE())

    def classification(
            self
    ) -> str:
        return "Encounter"

    def yaw(
            self
    ) -> Optional[float]:
        return None


class GITPlaceable(GITInstance):
    GFF_STRUCT_ID = 9

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        super().__init__(x, y, z)
        self.resref: ResRef = ResRef.from_blank()
        self.bearing: float = 0.0
        self.tweak_color: Optional[Color] = None

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        self.bearing += yaw

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return ResourceIdentifier(self.resref.get(), ResourceType.UTP)

    def blank(
            self
    ) -> Optional[bytes]:
        return bytes_utp(UTP())

    def classification(
            self
    ) -> str:
        return "Placeable"

    def yaw(
            self
    ) -> Optional[float]:
        return self.bearing


class GITSound(GITInstance):
    GFF_STRUCT_ID = 6

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        super().__init__(x, y, z)
        self.resref: ResRef = ResRef.from_blank()

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        raise ValueError("Sounds cannot be rotated.")

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return ResourceIdentifier(self.resref.get(), ResourceType.UTS)

    def blank(
            self
    ) -> Optional[bytes]:
        return bytes_uts(UTS())

    def classification(
            self
    ) -> str:
        return "Sound"

    def yaw(
            self
    ) -> Optional[float]:
        return None


class GITStore(GITInstance):
    GFF_STRUCT_ID = 11

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        super().__init__(x, y, z)
        self.resref: ResRef = ResRef.from_blank()
        self.bearing: float = 0.0

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        self.bearing += yaw

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return ResourceIdentifier(self.resref.get(), ResourceType.UTM)

    def blank(
            self
    ) -> Optional[bytes]:
        return bytes_utm(UTM())
    
    def classification(
            self
    ) -> str:
        return "Store"

    def yaw(
            self
    ) -> Optional[float]:
        return self.bearing


class GITTrigger(GITInstance):
    GFF_STRUCT_ID = 1
    GFF_GEOMETRY_STRUCT_ID = 3

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        super().__init__(x, y, z)
        self.resref: ResRef = ResRef.from_blank()
        self.geometry: Polygon3 = Polygon3()
        self.tag: str = ""
        self.linked_to: str = ""
        self.linked_to_flags: GITModuleLink = GITModuleLink.NoLink
        self.linked_to_module: ResRef = ResRef.from_blank()
        self.transition_destination: LocalizedString = LocalizedString.from_invalid()

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        raise ValueError("Triggers cannot be rotated.")

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return ResourceIdentifier(self.resref.get(), ResourceType.UTT)

    def blank(
            self
    ) -> Optional[bytes]:
        return bytes_utt(UTT())

    def classification(
            self
    ) -> str:
        return "Trigger"

    def yaw(
            self
    ) -> Optional[float]:
        return None


class GITTransitionTrigger(GITTrigger):
    def __init__(
            self
    ):
        super().__init__()
        self.linked_to: str = ""
        self.linked_to_flags: GITModuleLink = GITModuleLink.NoLink
        self.linked_to_module: ResRef = ResRef.from_blank()
        self.transition_destination: LocalizedString = LocalizedString.from_invalid()
        self.tag: str = ""


class GITWaypoint(GITInstance):
    GFF_STRUCT_ID = 5

    def __init__(
            self,
            x: float = 0.0,
            y: float = 0.0,
            z: float = 0.0
    ):
        super().__init__(x, y, z)
        self.resref: ResRef = ResRef.from_blank()
        self.tag: str = ""
        self.name: LocalizedString = LocalizedString.from_invalid()
        self.map_note: Optional[LocalizedString] = LocalizedString.from_invalid()
        self.map_note_enabled: bool = False
        self.has_map_note: bool = False
        self.bearing: float = 0.0

    def move(
            self,
            x: float,
            y: float,
            z: float
    ) -> None:
        self.position.x += x
        self.position.y += y
        self.position.z += z

    def rotate(
            self,
            yaw: float,
            pitch: float,
            roll: float
    ) -> None:
        self.bearing += yaw

    def identifier(
            self
    ) -> Optional[ResourceIdentifier]:
        return ResourceIdentifier(self.resref.get(), ResourceType.UTW)

    def blank(
            self
    ) -> Optional[bytes]:
        return bytes_utw(UTW())

    def classification(
            self
    ) -> str:
        return "Waypoint"

    def yaw(
            self
    ) -> Optional[float]:
        return self.bearing


def construct_git(
        gff: GFF
) -> GIT:
    git = GIT()

    root = gff.root
    properties_struct = root.acquire("AreaProperties", GFFStruct())
    git.ambient_volume = properties_struct.acquire("AmbientSndDayVol", 0)
    git.ambient_sound_id = properties_struct.acquire("AmbientSndDay", 0)
    git.env_audio = properties_struct.acquire("EnvAudio", 0)
    git.music_standard_id = properties_struct.acquire("MusicDay", 0)
    git.music_battle_id = properties_struct.acquire("MusicBattle", 0)
    git.music_delay = properties_struct.acquire("MusicDelay", 0)

    for camera_struct in gff.root.get_list("CameraList"):
        camera = GITCamera()
        git.cameras.append(camera)

        camera.camera_id = camera_struct.acquire("CameraID", 0)
        camera.fov = camera_struct.acquire("FieldOfView", 0.0)
        camera.height = camera_struct.acquire("Height", 0.0)
        camera.mic_range = camera_struct.acquire("MicRange", 0.0)
        camera.orientation = camera_struct.acquire("Orientation", Vector4.from_null())
        camera.position = camera_struct.acquire("Position", Vector3.from_null())
        camera.pitch = camera_struct.acquire("Pitch", 0.0)

    for creature_struct in gff.root.get_list("Creature List"):
        creature = GITCreature()
        git.creatures.append(creature)
        creature.resref = creature_struct.acquire("TemplateResRef", ResRef.from_blank())
        creature.position.x = creature_struct.acquire("XPosition", 0.0)
        creature.position.y = creature_struct.acquire("YPosition", 0.0)
        creature.position.z = creature_struct.acquire("ZPosition", 0.0)
        rot_x, rot_y = creature_struct.acquire("XOrientation", 0.0), creature_struct.acquire("YOrientation", 0.0)
        creature.bearing = Vector2(rot_x, rot_y).angle() - math.pi/2

    for door_struct in gff.root.get_list("Door List"):
        door = GITDoor()
        git.doors.append(door)
        door.bearing = door_struct.acquire("Bearing", 0.0)
        door.tag = door_struct.acquire("Tag", "")
        door.resref = door_struct.acquire("TemplateResRef", ResRef.from_blank())
        door.linked_to = door_struct.acquire("LinkedTo", "")
        door.linked_to_flags = GITModuleLink(door_struct.acquire("LinkedToFlags", 0))
        door.linked_to_module = door_struct.acquire("LinkedToModule", ResRef.from_blank())
        door.transition_destination = door_struct.acquire("TransitionDestin", LocalizedString.from_invalid())
        door.position.x = door_struct.acquire("X", 0.0)
        door.position.y = door_struct.acquire("Y", 0.0)
        door.position.z = door_struct.acquire("Z", 0.0)
        tweak_enabled = door_struct.acquire("UseTweakColor", 0)
        door.tweak_color = Color.from_bgr_integer(door_struct.acquire("TweakColor", 0)) if tweak_enabled else None

    for encounter_struct in gff.root.get_list("Encounter List"):
        x = encounter_struct.acquire("XPosition", 0.0)
        y = encounter_struct.acquire("YPosition", 0.0)
        z = encounter_struct.acquire("ZPosition", 0.0)

        encounter = GITEncounter()
        git.encounters.append(encounter)
        encounter.position = Vector3(x, y, z)
        encounter.resref = encounter_struct.acquire("TemplateResRef", ResRef.from_blank())

        for geometry_struct in encounter_struct.get_list("Geometry"):
            x = geometry_struct.acquire("X", 0.0)
            y = geometry_struct.acquire("Y", 0.0)
            z = geometry_struct.acquire("Z", 0.0)
            encounter.geometry.append(Vector3(x, y, z))

        for spawn_struct in encounter_struct.get_list("SpawnPointList"):
            x = spawn_struct.acquire("X", 0.0)
            y = spawn_struct.acquire("Y", 0.0)
            z = spawn_struct.acquire("Z", 0.0)
            spawn = GITEncounterSpawnPoint()
            spawn.position = Vector3(x, y, z)
            spawn.orientation = spawn_struct.acquire("Orientation", 0.0)
            encounter.spawn_points.append(spawn)

    for placeable_struct in gff.root.get_list("Placeable List"):
        placeable = GITPlaceable()
        git.placeables.append(placeable)

        placeable.resref = placeable_struct.acquire("TemplateResRef", ResRef.from_blank())
        placeable.position.x = placeable_struct.acquire("X", 0.0)
        placeable.position.y = placeable_struct.acquire("Y", 0.0)
        placeable.position.z = placeable_struct.acquire("Z", 0.0)
        placeable.bearing = placeable_struct.acquire("Bearing", 0.0)

        tweak_enabled = placeable_struct.acquire("UseTweakColor", 0)
        tweak_int = placeable_struct.acquire("TweakColor", 0)
        placeable.tweak_color = Color.from_bgr_integer(tweak_int) if tweak_enabled else None

    for sound_struct in gff.root.get_list("SoundList"):
        sound = GITSound()
        git.sounds.append(sound)

        sound.resref = sound_struct.acquire("TemplateResRef", ResRef.from_blank())
        sound.position.x = sound_struct.acquire("XPosition", 0.0)
        sound.position.y = sound_struct.acquire("YPosition", 0.0)
        sound.position.z = sound_struct.acquire("ZPosition", 0.0)

    for store_struct in gff.root.get_list("StoreList"):
        store = GITStore()
        git.stores.append(store)

        store.resref = store_struct.acquire("ResRef", ResRef.from_blank())
        store.position.x = store_struct.acquire("XPosition", 0.0)
        store.position.y = store_struct.acquire("YPosition", 0.0)
        store.position.z = store_struct.acquire("ZPosition", 0.0)

        rot_x, rot_y = store_struct.acquire("XOrientation", 0.0), store_struct.acquire("YOrientation", 0.0)
        store.bearing = Vector2(rot_x, rot_y).angle() - math.pi/2

    for trigger_struct in gff.root.get_list("TriggerList"):
        trigger = GITTrigger()
        git.triggers.append(trigger)

        trigger.resref = trigger_struct.acquire("TemplateResRef", ResRef.from_blank())
        trigger.position.x = trigger_struct.acquire("XPosition", 0.0)
        trigger.position.y = trigger_struct.acquire("YPosition", 0.0)
        trigger.position.z = trigger_struct.acquire("ZPosition", 0.0)
        trigger.tag = trigger_struct.acquire("Tag", "")
        trigger.linked_to = trigger_struct.acquire("LinkedTo", "")
        trigger.linked_to_flags = GITModuleLink(trigger_struct.acquire("LinkedToFlags", 0))
        trigger.linked_to_module = trigger_struct.acquire("LinkedToModule", ResRef.from_blank())
        trigger.transition_destination = trigger_struct.acquire("TransitionDestin", LocalizedString.from_invalid())

        for geometry_struct in trigger_struct.get_list("Geometry"):
            x = geometry_struct.acquire("PointX", 0.0)
            y = geometry_struct.acquire("PointY", 0.0)
            z = geometry_struct.acquire("PointZ", 0.0)
            trigger.geometry.append(Vector3(x, y, z))

    for waypoint_struct in gff.root.get_list("WaypointList"):
        waypoint = GITWaypoint()
        git.waypoints.append(waypoint)

        waypoint.name = waypoint_struct.acquire("LocalizedName", LocalizedString.from_invalid())
        waypoint.tag = waypoint_struct.acquire("Tag", "")
        waypoint.resref = waypoint_struct.acquire("TemplateResRef", ResRef.from_blank())
        waypoint.position.x = waypoint_struct.acquire("XPosition", 0.0)
        waypoint.position.y = waypoint_struct.acquire("YPosition", 0.0)
        waypoint.position.z = waypoint_struct.acquire("ZPosition", 0.0)

        waypoint.has_map_note = waypoint_struct.acquire("HasMapNote", 0)
        if waypoint.has_map_note:
            waypoint.map_note = waypoint_struct.acquire("MapNote", LocalizedString.from_invalid())
            waypoint.map_note_enabled = waypoint_struct.acquire("MapNoteEnabled", 0)

        rot_x, rot_y = waypoint_struct.acquire("XOrientation", 0.0), waypoint_struct.acquire("YOrientation", 0.0)
        waypoint.bearing = Vector2(rot_x, rot_y).angle() - math.pi/2

    return git


def dismantle_git(
        git: GIT,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.GIT)

    root = gff.root
    root.set_uint8("UseTemplates", 1)

    properties_struct = root.set_struct("AreaProperties", GFFStruct())
    properties_struct.set_int32("AmbientSndDayVol", git.ambient_volume)
    properties_struct.set_int32("AmbientSndDay", git.ambient_sound_id)
    properties_struct.set_int32("AmbientSndNitVol", git.ambient_volume)
    properties_struct.set_int32("AmbientSndNight", git.ambient_sound_id)
    properties_struct.set_int32("EnvAudio", git.env_audio)
    properties_struct.set_int32("MusicDay", git.music_standard_id)
    properties_struct.set_int32("MusicNight", git.music_standard_id)
    properties_struct.set_int32("MusicBattle", git.music_battle_id)
    properties_struct.set_int32("MusicDelay", git.music_delay)

    camera_list = root.set_list("CameraList", GFFList())
    for camera in git.cameras:
        camera_struct = camera_list.add(GITCamera.GFF_STRUCT_ID)
        camera_struct.set_int32("CameraID", camera.camera_id)
        camera_struct.set_single("FieldOfView", camera.fov)
        camera_struct.set_single("Height", camera.height)
        camera_struct.set_single("MicRange", camera.mic_range)
        camera_struct.set_vector4("Orientation", camera.orientation)
        camera_struct.set_vector3("Position", camera.position)
        camera_struct.set_single("Pitch", camera.pitch)

    creature_list = root.set_list("Creature List", GFFList())
    for creature in git.creatures:
        bearing = Vector2.from_angle(creature.bearing + math.pi/2)

        creature_struct = creature_list.add(GITCreature.GFF_STRUCT_ID)
        creature_struct.set_resref("TemplateResRef", creature.resref)
        creature_struct.set_single("XOrientation", bearing.x)
        creature_struct.set_single("YOrientation", bearing.y)
        creature_struct.set_single("XPosition", creature.position.x)
        creature_struct.set_single("YPosition", creature.position.y)
        creature_struct.set_single("ZPosition", creature.position.z)

    door_list = root.set_list("Door List", GFFList())
    for door in git.doors:
        door_struct = door_list.add(GITDoor.GFF_STRUCT_ID)
        door_struct.set_single("Bearing", door.bearing)
        door_struct.set_string("Tag", door.tag)
        door_struct.set_resref("TemplateResRef", door.resref)
        door_struct.set_string("LinkedTo", door.linked_to)
        door_struct.set_uint8("LinkedToFlags", door.linked_to_flags.value)
        door_struct.set_resref("LinkedToModule", door.linked_to_module)
        door_struct.set_locstring("TransitionDestin", door.transition_destination)
        door_struct.set_single("X", door.position.x)
        door_struct.set_single("Y", door.position.y)
        door_struct.set_single("Z", door.position.z)
        if game is Game.K2:
            tweak_color = door.tweak_color.bgr_integer() if door.tweak_color is not None else 0
            door_struct.set_uint32("TweakColor", tweak_color)
            door_struct.set_uint8("UseTweakColor", 0 if door.tweak_color is None else 1)

    encounter_list = root.set_list("Encounter List", GFFList())
    for encounter in git.encounters:
        encounter_struct = encounter_list.add(GITEncounter.GFF_STRUCT_ID)
        encounter_struct.set_resref("TemplateResRef", encounter.resref)
        encounter_struct.set_single("XPosition", encounter.position.x)
        encounter_struct.set_single("YPosition", encounter.position.y)
        encounter_struct.set_single("ZPosition", encounter.position.z)

        geometry_list = encounter_struct.set_list("Geometry", GFFList())
        for point in encounter.geometry:
            geometry_struct = geometry_list.add(GITEncounter.GFF_GEOMETRY_STRUCT_ID)
            geometry_struct.set_single("X", point.x)
            geometry_struct.set_single("Y", point.y)
            geometry_struct.set_single("Z", point.z)

        spawn_list = encounter_struct.set_list("SpawnPointList", GFFList())
        for spawn in encounter.spawn_points:
            spawn_struct = spawn_list.add(GITEncounter.GFF_SPAWN_STRUCT_ID)
            spawn_struct.set_single("Orientation", spawn.orientation)
            spawn_struct.set_single("X", spawn.x)
            spawn_struct.set_single("Y", spawn.y)
            spawn_struct.set_single("Z", spawn.z)

    placeable_list = root.set_list("Placeable List", GFFList())
    for placeable in git.placeables:
        placeable_struct = placeable_list.add(GITPlaceable.GFF_STRUCT_ID)
        placeable_struct.set_single("Bearing", placeable.bearing)
        placeable_struct.set_resref("TemplateResRef", placeable.resref)
        placeable_struct.set_single("X", placeable.position.x)
        placeable_struct.set_single("Y", placeable.position.y)
        placeable_struct.set_single("Z", placeable.position.z)
        if game is Game.K2:
            tweak_color = placeable.tweak_color.bgr_integer() if placeable.tweak_color is not None else 0
            placeable_struct.set_uint32("TweakColor", tweak_color)
            placeable_struct.set_uint8("UseTweakColor", 0 if placeable.tweak_color is None else 1)

    sound_list = root.set_list("SoundList", GFFList())
    for sound in git.sounds:
        sound_struct = sound_list.add(GITSound.GFF_STRUCT_ID)
        sound_struct.set_uint32("GeneratedType", 0)
        sound_struct.set_resref("TemplateResRef", sound.resref)
        sound_struct.set_single("XPosition", sound.position.x)
        sound_struct.set_single("YPosition", sound.position.y)
        sound_struct.set_single("ZPosition", sound.position.z)

    store_list = root.set_list("StoreList", GFFList())
    for store in git.stores:
        bearing = Vector2.from_angle(store.bearing + math.pi/2)

        store_struct = store_list.add(GITStore.GFF_STRUCT_ID)
        store_struct.set_resref("ResRef", store.resref)
        store_struct.set_single("XOrientation", bearing.x)
        store_struct.set_single("YOrientation", bearing.y)
        store_struct.set_single("XPosition", store.position.x)
        store_struct.set_single("YPosition", store.position.y)
        store_struct.set_single("ZPosition", store.position.z)

    trigger_list = root.set_list("TriggerList", GFFList())
    for trigger in git.triggers:
        trigger_struct = trigger_list.add(GITTrigger.GFF_STRUCT_ID)
        trigger_struct.set_resref("TemplateResRef", trigger.resref)
        trigger_struct.set_single("XPosition", trigger.position.x)
        trigger_struct.set_single("YPosition", trigger.position.y)
        trigger_struct.set_single("ZPosition", trigger.position.z)
        trigger_struct.set_single("XOrientation", 0.0)
        trigger_struct.set_single("YOrientation", 0.0)
        trigger_struct.set_single("ZOrientation", 0.0)

        trigger_struct.set_string("Tag", trigger.tag)
        trigger_struct.set_string("LinkedTo", trigger.linked_to)
        trigger_struct.set_uint8("LinkedToFlags", trigger.linked_to_flags.value)
        trigger_struct.set_resref("LinkedToModule", trigger.linked_to_module)
        trigger_struct.set_locstring("TransitionDestin", trigger.transition_destination)

        geometry_list = trigger_struct.set_list("Geometry", GFFList())
        for point in trigger.geometry:
            geometry_struct = geometry_list.add(GITTrigger.GFF_GEOMETRY_STRUCT_ID)
            geometry_struct.set_single("PointX", point.x)
            geometry_struct.set_single("PointY", point.y)
            geometry_struct.set_single("PointZ", point.z)

    waypoint_list = root.set_list("WaypointList", GFFList())
    for waypoint in git.waypoints:
        bearing = Vector2.from_angle(waypoint.bearing + math.pi/2)

        waypoint_struct = waypoint_list.add(GITWaypoint.GFF_STRUCT_ID)

        waypoint_struct.set_locstring("LocalizedName", waypoint.name)
        waypoint_struct.set_string("Tag", waypoint.tag)
        waypoint_struct.set_resref("TemplateResRef", waypoint.resref)
        waypoint_struct.set_single("XPosition", waypoint.position.x)
        waypoint_struct.set_single("YPosition", waypoint.position.y)
        waypoint_struct.set_single("ZPosition", waypoint.position.z)
        waypoint_struct.set_single("XOrientation", bearing.x)
        waypoint_struct.set_single("YOrientation", bearing.y)
        waypoint_struct.set_uint8("MapNoteEnabled", waypoint.map_note_enabled)
        waypoint_struct.set_uint8("HasMapNote", waypoint.has_map_note)
        waypoint_struct.set_locstring("MapNote", LocalizedString.from_invalid() if waypoint.map_note is None else waypoint.map_note)

    return gff


def read_git(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> GIT:
    gff = read_gff(source, offset, size)
    git = construct_git(gff)
    return git


def write_git(
        git: GIT,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_git(git, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_git(
        git: GIT,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_git(git, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from __future__ import annotations

from typing import List, Dict

from pykotor.common.language import LocalizedString
from pykotor.common.misc import ResRef, EquipmentSlot, InventoryItem, Game
from pykotor.resource.formats.gff import GFF, GFFList, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class UTC:
    """
    Stores creature data.

    Attributes:
        resref: "TemplateResRef" field.
        tag: "Tag" field.
        comment: "Comment" field.
        conversation: "Conversation" field.
        first_name: "FirstName" field.
        last_name: "LastName" field.
        subrace_id: "SubraceIndex" field.
        perception_id: "PerceptionRange" field.
        race_id: "Race" field.
        appearance_id: "Appearance_Type" field.
        gender_id: "Gender" field.
        faction_id: "FactionID" field.
        walkrate_id: "WalkRate" field.
        soundset_id: "SoundSetFile" field.
        portrait_id: "PortraitId" field.
        body_variation: "BodyVariation" field.
        texture_variation: "TextureVar" field.
        not_reorienting: "NotReorienting" field.
        party_interact: "PartyInteract" field.
        no_perm_death: "NoPermDeath" field.
        min1_hp: "Min1HP" field.
        plot: "Plot" field.
        interruptable: "Interruptable" field.
        is_pc: "IsPC" field.
        disarmable: "Disarmable" field.
        alignment: "GoodEvil" field.
        challenge_rating: "ChallengeRating" field.
        blindspot: "BlindSpot" field.
        multiplier_set: "MultiplierSet" field.
        natural_ac: "NaturalAC" field.
        reflex_bonus: "refbonus" field.
        willpower_bonus: "willbonus" field.
        fortitude_bonus: "fortbonus" field.
        strength: "Str" field.
        dexterity: "Dex" field.
        constitution: "Con" field.
        intelligence: "Int" field.
        wisdom: "Wis" field.
        charisma: "Cha" field.
        current_hp: "CurrentHitPoints" field.
        max_hp: "MaxHitPoints" field.
        hp: "HitPoints" field.
        fp: "CurrentForce" field.
        max_fp: "ForcePoints" field.
        on_end_dialog: "ScriptEndDialogu" field.
        on_blocked: "ScriptOnBlocked" field.
        on_heartbeat: "ScriptHeartbeat" field.
        on_notice: "ScriptOnNotice" field.
        on_spell: "ScriptSpellAt" field.
        on_attacked: "ScriptAttacked" field.
        on_damaged: "ScriptDamaged" field.
        on_disturbed: "ScriptDisturbed" field.
        on_end_round: "ScriptEndRound" field.
        on_dialog: "ScriptDialogue" field.
        on_spawn: "ScriptSpawn" field.
        on_rested: "ScriptRested" field.
        on_death: "ScriptDeath" field.
        on_user_defined: "ScriptUserDefine" field.

        ignore_cre_path: "IgnoreCrePath" field. KotOR 2 Only.
        hologram: "Hologram" field. KotOR 2 Only.

        palette_id: "PaletteID" field. Used in toolset only.

        bodybag_id: "BodyBag" field. Not used by the game engine.
        deity: "Deity" field. Not used by the game engine.
        description: "Description" field. Not used by the game engine.
        lawfulness: "LawfulChaotic" field. Not used by the game engine.
        phenotype_id: "Phenotype" field. Not used by the game engine.
        on_rested: "ScriptRested" field. Not used by the game engine.
        subrace_name: "Subrace" field. Not used by the game engine.
    """

    BINARY_TYPE = ResourceType.UTC

    def __init__(
            self
    ):
        self.resref: ResRef = ResRef.from_blank()
        self.conversation: ResRef = ResRef.from_blank()
        self.tag: str = ""
        self.comment: str = ""

        self.first_name: LocalizedString = LocalizedString.from_invalid()
        self.last_name: LocalizedString = LocalizedString.from_invalid()

        self.subrace_id: int = 0
        self.portrait_id: int = 0
        self.perception_id: int = 0
        self.race_id: int = 0
        self.appearance_id: int = 0
        self.gender_id: int = 0
        self.faction_id: int = 0
        self.walkrate_id: int = 0
        self.soundset_id: int = 0

        self.body_variation: int = 0
        self.texture_variation: int = 0

        self.not_reorienting: bool = False
        self.party_interact: bool = False
        self.no_perm_death: bool = False
        self.min1_hp: bool = False
        self.plot: bool = False
        self.interruptable: bool = False
        self.is_pc: bool = False  # ???
        self.disarmable: bool = False  # ???
        self.ignore_cre_path: bool = False  # KotOR 2 Only
        self.hologram: bool = False  # KotOR 2 Only

        self.alignment: int = 0
        self.challenge_rating: float = 0.0
        self.blindspot: float = 0.0  # KotOR 2 Only
        self.multiplier_set: int = 0  # KotOR 2 Only

        self.natural_ac: int = 0
        self.reflex_bonus: int = 0
        self.willpower_bonus: int = 0
        self.fortitude_bonus: int = 0

        self.current_hp: int = 0
        self.max_hp: int = 0
        self.hp: int = 0

        self.max_fp: int = 0
        self.fp: int = 0

        self.strength: int = 0
        self.dexterity: int = 0
        self.constitution: int = 0
        self.intelligence: int = 0
        self.wisdom: int = 0
        self.charisma: int = 0

        self.computer_use: int = 0
        self.demolitions: int = 0
        self.stealth: int = 0
        self.awareness: int = 0
        self.persuade: int = 0
        self.repair: int = 0
        self.security: int = 0
        self.treat_injury: int = 0

        self.on_end_dialog: ResRef = ResRef.from_blank()
        self.on_blocked: ResRef = ResRef.from_blank()
        self.on_heartbeat: ResRef = ResRef.from_blank()
        self.on_notice: ResRef = ResRef.from_blank()
        self.on_spell: ResRef = ResRef.from_blank()
        self.on_attacked: ResRef = ResRef.from_blank()
        self.on_damaged: ResRef = ResRef.from_blank()
        self.on_disturbed: ResRef = ResRef.from_blank()
        self.on_end_round: ResRef = ResRef.from_blank()
        self.on_dialog: ResRef = ResRef.from_blank()
        self.on_spawn: ResRef = ResRef.from_blank()
        self.on_rested: ResRef = ResRef.from_blank()
        self.on_death: ResRef = ResRef.from_blank()
        self.on_user_defined: ResRef = ResRef.from_blank()

        self.classes: List[UTCClass] = []
        self.feats: List[int] = []
        self.inventory: List[InventoryItem] = []
        self.equipment: Dict[EquipmentSlot, InventoryItem] = {}

        # Deprecated:
        self.palette_id: int = 0
        self.bodybag_id: int = 1
        self.deity: str = ""
        self.description: LocalizedString = LocalizedString.from_invalid()
        self.lawfulness: int = 0
        self.phenotype_id: int = 0
        self.on_rested: ResRef = ResRef.from_blank()
        self.subrace_name: str = ""


class UTCClass:
    def __init__(
            self,
            class_id: int,
            class_level: int = 0
    ):
        self.class_id: int = class_id
        self.class_level: int = class_level
        self.powers: List[int] = []

    def __eq__(
            self,
            other: UTCClass
    ):
        return self.class_id == other.class_id and self.class_level == self.class_level


def construct_utc(
        gff: GFF
) -> UTC:
    utc = UTC()

    root = gff.root
    utc.resref = root.acquire("TemplateResRef", ResRef.from_blank())
    utc.tag = root.acquire("Tag", "", str)
    utc.comment = root.acquire("Comment", "", str)
    utc.conversation = root.acquire("Conversation", ResRef.from_blank())

    utc.first_name = root.acquire("FirstName", LocalizedString.from_invalid())
    utc.last_name = root.acquire("LastName", LocalizedString.from_invalid())

    utc.subrace_id = root.acquire("SubraceIndex", 0)
    utc.perception_id = root.acquire("PerceptionRange", 0)
    utc.race_id = root.acquire("Race", 0)
    utc.appearance_id = root.acquire("Appearance_Type", 0)
    utc.gender_id = root.acquire("Gender", 0)
    utc.faction_id = root.acquire("FactionID", 0)
    utc.walkrate_id = root.acquire("WalkRate", 0)
    utc.soundset_id = root.acquire("SoundSetFile", 0)
    utc.portrait_id = root.acquire("PortraitId", 0)
    utc.palette_id = root.acquire("PaletteID", 0)

    utc.body_variation = root.acquire("BodyVariation", 0)
    utc.texture_variation = root.acquire("TextureVar", 0)

    utc.not_reorienting = bool(root.acquire("NotReorienting", 0))
    utc.party_interact = bool(root.acquire("PartyInteract", 0))
    utc.no_perm_death = bool(root.acquire("NoPermDeath", 0))
    utc.min1_hp = bool(root.acquire("Min1HP", 0))
    utc.plot = bool(root.acquire("Plot", 0))
    utc.interruptable = bool(root.acquire("Interruptable", 0))
    utc.is_pc = bool(root.acquire("IsPC", 0))
    utc.disarmable = bool(root.acquire("Disarmable", 0))
    utc.ignore_cre_path = bool(root.acquire("IgnoreCrePath", 0))
    utc.hologram = bool(root.acquire("Hologram", 0))

    utc.alignment = root.acquire("GoodEvil", 0)
    utc.challenge_rating = root.acquire("ChallengeRating", 0.0)
    utc.blindspot = root.acquire("BlindSpot", 0.0)
    utc.multiplier_set = root.acquire("MultiplierSet", 0)

    utc.natural_ac = root.acquire("NaturalAC", 0)
    utc.reflex_bonus = root.acquire("refbonus", 0)
    utc.willpower_bonus = root.acquire("willbonus", 0)
    utc.fortitude_bonus = root.acquire("fortbonus", 0)

    utc.strength = root.acquire("Str", 0)
    utc.dexterity = root.acquire("Dex", 0)
    utc.constitution = root.acquire("Con", 0)
    utc.intelligence = root.acquire("Int", 0)
    utc.wisdom = root.acquire("Wis", 0)
    utc.charisma = root.acquire("Cha", 0)

    utc.current_hp = root.acquire("CurrentHitPoints", 0)
    utc.max_hp = root.acquire("MaxHitPoints", 0)
    utc.hp = root.acquire("HitPoints", 0)
    utc.max_fp = root.acquire("ForcePoints", 0)
    utc.fp = root.acquire("CurrentForce", 0)

    utc.on_end_dialog = root.acquire("ScriptEndDialogu", ResRef.from_blank())
    utc.on_blocked = root.acquire("ScriptOnBlocked", ResRef.from_blank())
    utc.on_heartbeat = root.acquire("ScriptHeartbeat", ResRef.from_blank())
    utc.on_notice = root.acquire("ScriptOnNotice", ResRef.from_blank())
    utc.on_spell = root.acquire("ScriptSpellAt", ResRef.from_blank())
    utc.on_attacked = root.acquire("ScriptAttacked", ResRef.from_blank())
    utc.on_damaged = root.acquire("ScriptDamaged", ResRef.from_blank())
    utc.on_disturbed = root.acquire("ScriptDisturbed", ResRef.from_blank())
    utc.on_end_round = root.acquire("ScriptEndRound", ResRef.from_blank())
    utc.on_dialog = root.acquire("ScriptDialogue", ResRef.from_blank())
    utc.on_spawn = root.acquire("ScriptSpawn", ResRef.from_blank())
    utc.on_rested = root.acquire("ScriptRested", ResRef.from_blank())
    utc.on_death = root.acquire("ScriptDeath", ResRef.from_blank())
    utc.on_user_defined = root.acquire("ScriptUserDefine", ResRef.from_blank())

    skill_list: GFFList = root.acquire("SkillList", GFFList())
    utc.computer_use = skill_list.at(0).acquire("Rank", 0)
    utc.demolitions = skill_list.at(1).acquire("Rank", 0)
    utc.stealth = skill_list.at(2).acquire("Rank", 0)
    utc.awareness = skill_list.at(3).acquire("Rank", 0)
    utc.persuade = skill_list.at(4).acquire("Rank", 0)
    utc.repair = skill_list.at(5).acquire("Rank", 0)
    utc.security = skill_list.at(6).acquire("Rank", 0)
    utc.treat_injury = skill_list.at(7).acquire("Rank", 0)

    class_list: GFFList = root.acquire("ClassList", GFFList())
    for class_struct in class_list:
        class_id = class_struct.acquire("Class", 0)
        class_level = class_struct.acquire("ClassLevel", 0)
        utc_class = UTCClass(class_id, class_level)

        power_list: GFFList = class_struct.acquire("KnownList0", GFFList())
        for power_struct in power_list:
            utc_class.powers.append(power_struct.acquire("Spell", 0))

        utc.classes.append(utc_class)

    feat_list: GFFList = root.acquire("FeatList", GFFList())
    for feat_struct in feat_list:
        utc.feats.append(feat_struct.acquire("Feat", 0))

    equipment_list: GFFList = root.acquire("Equip_ItemList", GFFList())
    for equipment_struct in equipment_list:
        slot = EquipmentSlot(equipment_struct.struct_id)
        resref = equipment_struct.acquire("EquippedRes", ResRef.from_blank())
        droppable = bool(equipment_struct.acquire("Dropable", 0))
        utc.equipment[slot] = InventoryItem(resref, droppable)

    item_list: GFFList = root.acquire("ItemList", GFFList())
    for item_struct in item_list:
        resref = item_struct.acquire("InventoryRes", ResRef.from_blank())
        droppable = bool(item_struct.acquire("Dropable", 0))
        utc.inventory.append(InventoryItem(resref, droppable))

    return utc


def dismantle_utc(
        utc: UTC,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.UTC)

    root = gff.root
    root.set_resref("TemplateResRef", utc.resref)
    root.set_string("Tag", utc.tag)
    root.set_string("Comment", utc.comment)
    root.set_resref("Conversation", utc.conversation)

    root.set_locstring("FirstName", utc.first_name)
    root.set_locstring("LastName", utc.last_name)

    root.set_uint8("SubraceIndex", utc.subrace_id)
    root.set_uint8("PerceptionRange", utc.perception_id)
    root.set_uint8("Race", utc.race_id)
    root.set_uint16("Appearance_Type", utc.appearance_id)
    root.set_uint8("Gender", utc.gender_id)
    root.set_uint16("FactionID", utc.faction_id)
    root.set_int32("WalkRate", utc.walkrate_id)
    root.set_uint16("SoundSetFile", utc.soundset_id)
    root.set_uint16("PortraitId", utc.portrait_id)

    root.set_uint8("BodyVariation", utc.body_variation)
    root.set_uint8("TextureVar", utc.texture_variation)

    root.set_uint8("NotReorienting", utc.not_reorienting)
    root.set_uint8("PartyInteract", utc.party_interact)
    root.set_uint8("NoPermDeath", utc.no_perm_death)
    root.set_uint8("Min1HP", utc.min1_hp)
    root.set_uint8("Plot", utc.plot)
    root.set_uint8("Interruptable", utc.interruptable)
    root.set_uint8("IsPC", utc.is_pc)
    root.set_uint8("Disarmable", utc.disarmable)

    root.set_uint8("GoodEvil", utc.alignment)
    root.set_single("ChallengeRating", utc.challenge_rating)

    root.set_uint8("NaturalAC", utc.natural_ac)
    root.set_int16("refbonus", utc.reflex_bonus)
    root.set_int16("willbonus", utc.willpower_bonus)
    root.set_int16("fortbonus", utc.fortitude_bonus)

    root.set_uint8("Str", utc.strength)
    root.set_uint8("Dex", utc.dexterity)
    root.set_uint8("Con", utc.constitution)
    root.set_uint8("Int", utc.intelligence)
    root.set_uint8("Wis", utc.wisdom)
    root.set_uint8("Cha", utc.charisma)

    root.set_uint8("CurrentHitPoints", utc.current_hp)
    root.set_uint8("MaxHitPoints", utc.max_hp)
    root.set_uint8("HitPoints", utc.hp)
    root.set_uint8("CurrentForce", utc.fp)
    root.set_uint8("ForcePoints", utc.max_fp)

    root.set_resref("ScriptEndDialogu", utc.on_end_dialog)
    root.set_resref("ScriptOnBlocked", utc.on_blocked)
    root.set_resref("ScriptHeartbeat", utc.on_heartbeat)
    root.set_resref("ScriptOnNotice", utc.on_notice)
    root.set_resref("ScriptSpellAt", utc.on_spell)
    root.set_resref("ScriptAttacked", utc.on_attacked)
    root.set_resref("ScriptDamaged", utc.on_damaged)
    root.set_resref("ScriptDisturbed", utc.on_disturbed)
    root.set_resref("ScriptEndRound", utc.on_end_round)
    root.set_resref("ScriptDialogue", utc.on_dialog)
    root.set_resref("ScriptSpawn", utc.on_spawn)
    root.set_resref("ScriptRested", utc.on_rested)
    root.set_resref("ScriptDeath", utc.on_death)
    root.set_resref("ScriptUserDefine", utc.on_user_defined)

    root.set_uint8("PaletteID", utc.palette_id)

    skill_list = root.set_list("SkillList", GFFList())
    skill_list.add(0).set_uint8("Rank", utc.computer_use)
    skill_list.add(0).set_uint8("Rank", utc.demolitions)
    skill_list.add(0).set_uint8("Rank", utc.stealth)
    skill_list.add(0).set_uint8("Rank", utc.awareness)
    skill_list.add(0).set_uint8("Rank", utc.persuade)
    skill_list.add(0).set_uint8("Rank", utc.repair)
    skill_list.add(0).set_uint8("Rank", utc.security)
    skill_list.add(0).set_uint8("Rank", utc.treat_injury)

    class_list = root.set_list("ClassList", GFFList())
    for utc_class in utc.classes:
        class_struct = class_list.add(2)
        class_struct.set_int32("Class", utc_class.class_id)
        class_struct.set_int32("ClassLevel", utc_class.class_level)
        power_list = class_struct.set_list("KnownList0", GFFList())
        for power in utc_class.powers:
            power_struct = power_list.add(3)
            power_struct.set_uint16("Spell", power)
            power_struct.set_uint8("SpellFlags", 1)
            power_struct.set_uint8("SpellMetaMagic", 0)

    feat_list = root.set_list("FeatList", GFFList())
    for feat in utc.feats:
        feat_list.add(1).set_uint16("Feat", feat)

    equipment_list = root.set_list("Equip_ItemList", GFFList())
    for slot, item in utc.equipment.items():
        equipment_struct = equipment_list.add(slot.value)
        equipment_struct.set_resref("EquippedRes", item.resref)
        if item.droppable:
            equipment_struct.set_uint8("Dropable", True)

    item_list = root.set_list("ItemList", GFFList())
    for i, item in enumerate(utc.inventory):
        item_struct = item_list.add(i)
        item_struct.set_resref("InventoryRes", item.resref)
        item_struct.set_uint16("Repos_PosX", i)
        item_struct.set_uint16("Repos_Posy", 0)
        if item.droppable:
            item_struct.set_uint8("Dropable", True)

    if game == Game.K2:
        root.set_single("BlindSpot", utc.blindspot)
        root.set_uint8("MultiplierSet", utc.multiplier_set)
        root.set_uint8("IgnoreCrePath", utc.ignore_cre_path)
        root.set_uint8("Hologram", utc.hologram)

    if use_deprecated:
        root.set_uint8("BodyBag", utc.bodybag_id)
        root.set_string("Deity", utc.deity)
        root.set_locstring("Description", utc.description)
        root.set_uint8("LawfulChaotic", utc.lawfulness)
        root.set_int32("Phenotype", utc.phenotype_id)
        root.set_resref("ScriptRested", utc.on_rested)
        root.set_string("Subrace", utc.subrace_name)
        root.set_list("SpecAbilityList", GFFList())
        root.set_list("TemplateList", GFFList())

    return gff


def read_utc(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> UTC:
    gff = read_gff(source, offset, size)
    utc = construct_utc(gff)
    return utc


def write_utc(
        utc: UTC,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_utc(utc, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_utc(
        utc: UTC,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_utc(utc, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from __future__ import annotations

import base64
import io
from typing import Optional, Any
from xml.etree import ElementTree

from pykotor.common.geometry import Vector4, Vector3
from pykotor.common.language import LocalizedString
from pykotor.common.misc import ResRef
from pykotor.resource.formats.gff.gff_data import GFF, GFFStruct, GFFFieldType, GFFList
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter, autoclose


class GFFXMLReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = None
    ):
        super().__init__(source, offset, size)
        self._gff: Optional[GFF] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> GFF:
        self._gff = GFF()

        data = self._reader.read_bytes(self._reader.size()).decode()
        xml_root = ElementTree.parse(io.StringIO(data)).getroot().find('struct')
        self._load_struct(self._gff.root, xml_root)

        return self._gff

    def _load_struct(
            self,
            gff_struct: GFFStruct,
            xml_struct: ElementTree.Element
    ):
        gff_struct.struct_id = int(xml_struct.get("id"))

        for xml_field in xml_struct:
            self._load_field(gff_struct, xml_field)

    def _load_field(
            self,
            gff_struct: GFFStruct,
            xml_field: ElementTree.Element
    ):
        label = xml_field.get("label")

        if xml_field.tag == "byte":
            gff_struct.set_uint8(label, int(xml_field.text))
        elif xml_field.tag == "char":
            gff_struct.set_int8(label, int(xml_field.text))
        elif xml_field.tag == "uint16":
            gff_struct.set_uint16(label, int(xml_field.text))
        elif xml_field.tag == "sint16":
            gff_struct.set_int16(label, int(xml_field.text))
        elif xml_field.tag == "uint32":
            gff_struct.set_uint32(label, int(xml_field.text))
        elif xml_field.tag == "sint32":
            gff_struct.set_int32(label, int(xml_field.text))
        elif xml_field.tag == "uint64":
            gff_struct.set_uint64(label, int(xml_field.text))
        elif xml_field.tag == "sint65":
            gff_struct.set_int64(label, int(xml_field.text))
        elif xml_field.tag == "float":
            gff_struct.set_single(label, float(xml_field.text))
        elif xml_field.tag == "double":
            gff_struct.set_double(label, float(xml_field.text))
        elif xml_field.tag == "exostring":
            gff_struct.set_string(label, xml_field.text)
        elif xml_field.tag == "resref":
            gff_struct.set_resref(label, ResRef(xml_field.text))
        elif xml_field.tag == "locstring":
            locstring = LocalizedString(-1)
            locstring.stringref = -1 if xml_field.get("strref") == "4294967295" else int(xml_field.get("strref"))
            for substring in xml_field:
                language, gender = LocalizedString.substring_pair(int(substring.get("language")))
                locstring.set(language, gender, substring.text)
            gff_struct.set_locstring(label, locstring)
        elif xml_field.tag == "data":
            data = base64.b64decode(xml_field.text)
            gff_struct.set_binary(label, data)
        elif xml_field.tag == "orientation":
            coords = xml_field.findall('double')
            v4 = Vector4(float(coords[0].text), float(coords[1].text), float(coords[2].text), float(coords[3].text))
            gff_struct.set_vector4(label, v4)
        elif xml_field.tag == "vector":
            coords = xml_field.findall('double')
            v3 = Vector3(float(coords[0].text), float(coords[1].text), float(coords[2].text))
            gff_struct.set_vector3(label, v3)
        elif xml_field.tag == "struct":
            child_struct = GFFStruct()
            self._load_struct(child_struct, xml_field)
            gff_struct.set_struct(label, child_struct)
        elif xml_field.tag == "list":
            gff_list = GFFList()
            for xml_struct in xml_field:
                gff_list.add(0)
                child_struct = gff_list.at(len(gff_list) - 1)
                self._load_struct(child_struct, xml_struct)
            gff_struct.set_list(label, gff_list)


class GFFXMLWriter(ResourceWriter):
    def __init__(
            self,
            gff: GFF,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self.xml_root: ElementTree.Element = ElementTree.Element("xml")
        self.gff: GFF = gff

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        self.xml_root.tag = "gff3"

        xml_struct = ElementTree.Element("struct")
        self.xml_root.append(xml_struct)
        self._build_struct(self.gff.root, xml_struct)

        ElementTree.indent(self.xml_root)
        self._writer.write_bytes(ElementTree.tostring(self.xml_root))

    def _build_struct(
            self,
            gff_struct: GFFStruct,
            xml_struct: ElementTree.Element
    ):
        xml_struct.set('id', str(gff_struct.struct_id))

        for label, field_type, value in gff_struct:
            self._build_field(label, value, field_type, xml_struct)

    def _build_field(
            self,
            label: str,
            value: Any,
            field_type: GFFFieldType,
            xml_struct: ElementTree.Element
    ):
        xml_field = ElementTree.Element("")
        xml_field.set("label", label)
        xml_struct.append(xml_field)

        if field_type == GFFFieldType.UInt8:
            xml_field.tag = "byte"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.Int8:
            xml_field.tag = "char"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.UInt16:
            xml_field.tag = "uint16"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.Int16:
            xml_field.tag = "sint16"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.UInt32:
            xml_field.tag = "uint32"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.Int32:
            xml_field.tag = "sint32"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.UInt64:
            xml_field.tag = "uint64"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.Int64:
            xml_field.tag = "sint64"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.Single:
            xml_field.tag = "float"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.Double:
            xml_field.tag = "double"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.String:
            xml_field.tag = "exostring"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.ResRef:
            xml_field.tag = "resref"
            xml_field.text = str(value)
        elif field_type == GFFFieldType.LocalizedString:
            xml_field.tag = "locstring"
            xml_field.set("strref", str(value.stringref))
            for language, gender, string in value:
                subelement = ElementTree.Element("string")
                subelement.set("language", str(LocalizedString.substring_id(language, gender)))
                subelement.text = string
                xml_field.append(subelement)
        elif field_type == GFFFieldType.Binary:
            xml_field.tag = "data"
            xml_field.text = base64.b64encode(value).decode()
        elif field_type == GFFFieldType.Vector4:
            xml_field.tag = "orientation"
            x_element = ElementTree.Element("double")
            x_element.text = str(value.x)
            y_element = ElementTree.Element("double")
            y_element.text = str(value.y)
            z_element = ElementTree.Element("double")
            z_element.text = str(value.z)
            w_element = ElementTree.Element("double")
            w_element.text = str(value.w)
            xml_field.extend([x_element, y_element, z_element, w_element])
        elif field_type == GFFFieldType.Vector3:
            xml_field.tag = "vector"
            x_element = ElementTree.Element("double")
            x_element.text = str(value.x)
            y_element = ElementTree.Element("double")
            y_element.text = str(value.y)
            z_element = ElementTree.Element("double")
            z_element.text = str(value.z)
            xml_field.extend([x_element, y_element, z_element])
        elif field_type == GFFFieldType.Struct:
            xml_field.tag = "struct"
            self._build_struct(value, xml_field)
        elif field_type == GFFFieldType.List:
            xml_field.tag = "list"
            for gff_struct in value:
                subelement = ElementTree.Element("struct")
                xml_field.append(subelement)
                self._build_struct(gff_struct, subelement)



from __future__ import annotations

from contextlib import suppress
from typing import List, Optional

from pykotor.common.geometry import Vector2
from pykotor.common.misc import Game
from pykotor.resource.formats.gff import GFF, GFFList, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class PTH:
    """
    Stores the path data for a module.
    """

    BINARY_TYPE = ResourceType.PTH

    def __init__(
            self
    ):
        self._points: List[Vector2] = []
        self._connections: List[PTHEdge] = []

    def __iter__(
            self
    ):
        for point in self._points:
            yield point

    def __len__(
            self
    ):
        return len(self._points)

    def __getitem__(
            self,
            item: int
    ):
        return self._points[int]

    def add(
            self,
            x: float,
            y: float
    ) -> int:
        self._points.append(Vector2(x, y))
        return len(self._points) - 1

    def remove(
            self,
            index: int
    ) -> None:
        self._points.pop(index)

        self._connections = [x for x in self._connections if x.source != index and x.target != index]

        for i, connection in enumerate(self._connections):
            connection.source = connection.source - 1 if connection.source > index else connection.source
            connection.target = connection.target - 1 if connection.target > index else connection.target

    def get(
            self,
            index: int
    ) -> Optional[Vector2]:
        with suppress(Exception):
            return self._points[index]
        return None

    def find(
            self,
            point: Vector2
    ) -> Optional[int]:
        return next([x for x in self._points if x == point], None)

    def connect(
            self,
            source: int,
            target: int
    ) -> None:
        self._connections.append(PTHEdge(source, target))

    def disconnect(
            self,
            source: int,
            target: int
    ) -> None:
        self._connections.remove(PTHEdge(source, target))

    def is_connected(
            self,
            source: int,
            target: int
    ) -> bool:
        return any(x for x in self._connections if x == PTHEdge(source, target))

    def outgoing(
            self,
            source: int
    ) -> List[PTHEdge]:
        connections = []
        for connection in self._connections:
            if connection.source == source:
                connections.append(connection)
        return connections

    def incoming(
            self,
            target: int
    ) -> List[PTHEdge]:
        connections = []
        for connection in self._connections:
            if connection.target == target:
                connections.append(connection)
        return connections


class PTHEdge:
    def __init__(
            self,
            source: int,
            target: int
    ):
        self.source = source
        self.target = target

    def __eq__(
            self,
            other: PTHEdge
    ):
        if not isinstance(other, PTHEdge):
            raise NotImplemented

        return self.source == other.source and self.target == other.target


def construct_pth(
        gff: GFF
) -> PTH:
    pth = PTH()

    connections_list = gff.root.acquire("Path_Conections", GFFList())

    for point_struct in gff.root.acquire("Path_Points", GFFList()):
        connections = point_struct.acquire("Conections", 0)
        first_connection = point_struct.acquire("First_Conection", 0)
        x = point_struct.acquire("X", 0.0)
        y = point_struct.acquire("Y", 0.0)

        source = pth.add(x, y)

        for i in range(first_connection, first_connection + connections):
            target = connections_list.at(i).acquire("Destination", 0)
            pth.connect(source, target)

    return pth


def dismantle_pth(
        pth: PTH,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.PTH)

    connections_list = gff.root.set_list("Path_Conections", GFFList())
    points_list = gff.root.set_list("Path_Points", GFFList())

    for i, point in enumerate(pth):
        outgoings = pth.outgoing(i)

        point_struct = points_list.add(2)
        point_struct.set_uint32("Conections", len(outgoings))
        point_struct.set_uint32("First_Conection", len(connections_list))
        point_struct.set_single("X", point.x)
        point_struct.set_single("Y", point.y)

        for outgoing in outgoings:
            connection_struct = connections_list.add(3)
            connection_struct.set_uint32("Destination", outgoing.target)

    return gff


def read_pth(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> PTH:
    gff = read_gff(source, offset, size)
    pth = construct_pth(gff)
    return pth


def write_pth(
        pth: PTH,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_pth(pth, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_pth(
        pth: PTH,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_pth(pth, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from pykotor.resource.formats.tlk import TLK


class DiffTLK:
    def __init__(self, old: TLK, new: TLK):
        self.old: TLK = old
        self.new: TLK = new

    def is_same(self) -> bool:
        if len(self.old) != len(self.new):
            print("TLK row counts do not match")
            return False

        for stringref, entry in self.old:
            if entry.text != self.new.get(stringref).text:
                print("Entry text does not match:", stringref)
                return False
            if entry.voiceover != self.new.get(stringref).voiceover:
                print("Entry voiceover does not match:", stringref)
                return False

        return True



from unittest import TestCase

from pykotor.common.misc import ResRef

from pykotor.common.geometry import Vector3, Vector4
from pykotor.common.language import LocalizedString, Language, Gender
from pykotor.resource.formats.gff import GFF, GFFList, GFFFieldType
from pykotor.resource.formats.ssf import SSF, SSFSound
from pykotor.resource.formats.tlk import TLK
from pykotor.resource.formats.twoda import TwoDA
from pykotor.tslpatcher.mods.tlk import ModificationsTLK, ModifyTLK
from pykotor.tslpatcher.mods.gff import ModificationsGFF, ModifyFieldGFF, AddFieldGFF, AddStructToListGFF, \
    LocalizedStringDelta, FieldValueConstant, FieldValue2DAMemory, FieldValueTLKMemory
from pykotor.tslpatcher.memory import PatcherMemory, NoTokenUsage, TokenUsage2DA, TokenUsageTLK
from pykotor.tslpatcher.mods.ssf import ModificationsSSF, ModifySSF
from pykotor.tslpatcher.mods.twoda import Modifications2DA, ChangeRow2DA, Target, TargetType, AddRow2DA, CopyRow2DA, \
    AddColumn2DA, RowValueConstant, RowValueTLKMemory, RowValue2DAMemory, RowValueHigh, RowValueRowIndex, \
    RowValueRowLabel, RowValueRowCell


# TODO Error, Warning tracking

class TestManipulateTLK(TestCase):
    def test_apply(self):
        memory = PatcherMemory()

        config = ModificationsTLK()
        config.modifiers.append(ModifyTLK(0, "Append2", ResRef.from_blank()))
        config.modifiers.append(ModifyTLK(1, "Append1", ResRef.from_blank()))

        dialog_tlk = TLK()
        dialog_tlk.add("Old1")
        dialog_tlk.add("Old2")

        config.apply(dialog_tlk, memory)

        self.assertEqual(4, len(dialog_tlk))
        self.assertEqual("Append2", dialog_tlk.get(2).text)
        self.assertEqual("Append1", dialog_tlk.get(3).text)

        self.assertEqual(2, memory.memory_str[0])
        self.assertEqual(3, memory.memory_str[1])

        # [Dialog] [Append] [Token] [Text]
        # 0        -        -       Old1
        # 1        -        -       Old2
        # 2        1        0       Append2
        # 3        0        1       Append1


class TestManipulate2DA(TestCase):
    # region Change Row
    def test_change_existing_rowindex(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 1), {"Col1": RowValueConstant("X")}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "X"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f"], twoda.get_column("Col3"))

    def test_change_existing_rowlabel(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_LABEL, "1"), {"Col1": RowValueConstant("X")}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "X"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f"], twoda.get_column("Col3"))

    def test_change_existing_labelindex(self):
        twoda = TwoDA(["label", "Col2", "Col3"])
        twoda.add_row("0", {"label": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"label": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.LABEL_COLUMN, "d"), {"Col2": RowValueConstant("X")}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "d"], twoda.get_column("label"))
        self.assertEqual(["b", "X"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f"], twoda.get_column("Col3"))

    def test_change_assign_tlkmemory(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        memory.memory_str[0] = 0
        memory.memory_str[1] = 1
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 0), {"Col1": RowValueTLKMemory(0)}))
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 1), {"Col1": RowValueTLKMemory(1)}))
        config.apply(twoda, memory)

        self.assertEqual(["0", "1"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f"], twoda.get_column("Col3"))

    def test_change_assign_2damemory(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        memory.memory_2da[0] = "mem0"
        memory.memory_2da[1] = "mem1"
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 0), {"Col1": RowValue2DAMemory(0)}))
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 1), {"Col1": RowValue2DAMemory(1)}))
        config.apply(twoda, memory)

        self.assertEqual(["mem0", "mem1"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f"], twoda.get_column("Col3"))

    def test_change_assign_high(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": " ", "Col2": "3", "Col3": "5"})
        twoda.add_row("1", {"Col1": "2", "Col2": "4", "Col3": "6"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 0), {"Col1": RowValueHigh("Col1")}))
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 0), {"Col2": RowValueHigh("Col2")}))
        config.apply(twoda, memory)

        self.assertEqual(["3", "2"], twoda.get_column("Col1"))
        self.assertEqual(["5", "4"], twoda.get_column("Col2"))
        self.assertEqual(["5", "6"], twoda.get_column("Col3"))

    def test_set_2damemory_rowindex(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 1), {}, store_2da={5: RowValueRowIndex()}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "d"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f"], twoda.get_column("Col3"))
        self.assertEqual("1", memory.memory_2da[5])

    def test_set_2damemory_rowlabel(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("r1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 1), {}, store_2da={5: RowValueRowLabel()}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "d"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f"], twoda.get_column("Col3"))
        self.assertEqual("r1", memory.memory_2da[5])

    def test_set_2damemory_columnlabel(self):
        twoda = TwoDA(["label", "Col2", "Col3"])
        twoda.add_row("0", {"label": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"label": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(ChangeRow2DA("", Target(TargetType.ROW_INDEX, 1), {}, store_2da={5: RowValueRowCell("label")}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "d"], twoda.get_column("label"))
        self.assertEqual(["b", "e"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f"], twoda.get_column("Col3"))
        self.assertEqual("d", memory.memory_2da[5])
    # endregion

    # region Add Row
    def test_add_rowlabel_use_maxrowlabel(self):
        twoda = TwoDA(["Col1"])
        twoda.add_row("0", {})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", None, None, {}))
        config.modifiers.append(AddRow2DA("", None, None, {}))
        config.apply(twoda, memory)

        self.assertEqual(3, twoda.get_height())
        self.assertEqual("0", twoda.get_label(0))
        self.assertEqual("1", twoda.get_label(1))
        self.assertEqual("2", twoda.get_label(2))

    def test_add_rowlabel_use_constant(self):
        twoda = TwoDA(["Col1"])

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", None, "r1", {}))
        config.apply(twoda, memory)

        self.assertEqual(1, twoda.get_height())
        self.assertEqual("r1", twoda.get_label(0))

    def test_add_rowlabel_existing(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "123", "Col2": "456"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", "Col1", None, {"Col1": RowValueConstant("123"), "Col2": RowValueConstant("ABC")}))
        config.apply(twoda, memory)

        self.assertEqual(1, twoda.get_height())
        self.assertEqual("0", twoda.get_label(0))

    def test_add_exclusive_notexists(self):
        """Exclusive column is specified and the value in the new row is unique. Add a new row."""
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", "Col1", "2", {"Col1": RowValueConstant("g"), "Col2": RowValueConstant("h"), "Col3": RowValueConstant("i")}))
        config.apply(twoda, memory)

        self.assertEqual(3, twoda.get_height())
        self.assertEqual("2", twoda.get_label(2))
        self.assertEqual(["a", "d", "g"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e", "h"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f", "i"], twoda.get_column("Col3"))

    def test_add_exclusive_exists(self):
        """Exclusive column is specified but the value in the new row is already used. Edit the existing row."""
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})
        twoda.add_row("2", {"Col1": "g", "Col2": "h", "Col3": "i"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", "Col1", "3", {"Col1": RowValueConstant("g"), "Col2": RowValueConstant("X"), "Col3": RowValueConstant("Y")}))
        config.apply(twoda, memory)

        self.assertEqual(3, twoda.get_height())
        self.assertEqual(["a", "d", "g"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e", "X"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f", "Y"], twoda.get_column("Col3"))

    def test_add_exclusive_badcolumn(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", "Col4", "2", {}))
        # config.apply(twoda, memory)
        # TODO

    def test_add_exclusive_none(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "d", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", "", "2", {"Col1": RowValueConstant("g"), "Col2": RowValueConstant("h"), "Col3": RowValueConstant("i")}))
        config.modifiers.append(AddRow2DA("", None, "3", {"Col1": RowValueConstant("j"), "Col2": RowValueConstant("k"), "Col3": RowValueConstant("l")}))
        config.apply(twoda, memory)

        self.assertEqual(4, twoda.get_height())
        self.assertEqual(["a", "d", "g", "j"], twoda.get_column("Col1"))
        self.assertEqual(["b", "e", "h", "k"], twoda.get_column("Col2"))
        self.assertEqual(["c", "f", "i", "l"], twoda.get_column("Col3"))

    def test_add_assign_high(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "1", "Col2": "b", "Col3": "c"})
        twoda.add_row("1", {"Col1": "2", "Col2": "e", "Col3": "f"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", "", "2", {"Col1": RowValueHigh("Col1")}))
        config.apply(twoda, memory)

        self.assertEqual(["1", "2", "3"], twoda.get_column("Col1"))

    def test_add_assign_tlkmemory(self):
        twoda = TwoDA(["Col1"])

        memory = PatcherMemory()
        memory.memory_str[0] = 5
        memory.memory_str[1] = 6

        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", None, "0", {"Col1": RowValueTLKMemory(0)}))
        config.modifiers.append(AddRow2DA("", None, "1", {"Col1": RowValueTLKMemory(1)}))
        config.apply(twoda, memory)

        self.assertEqual(["5", "6"], twoda.get_column("Col1"))

    def test_add_assign_2damemory(self):
        twoda = TwoDA(["Col1"])

        memory = PatcherMemory()
        memory.memory_2da[0] = "5"
        memory.memory_2da[1] = "6"

        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", None, "0", {"Col1":  RowValue2DAMemory(0)}))
        config.modifiers.append(AddRow2DA("", None, "1", {"Col1":  RowValue2DAMemory(1)}))
        config.apply(twoda, memory)

        self.assertEqual(["5", "6"], twoda.get_column("Col1"))

    def test_add_2damemory_rowindex(self):
        twoda = TwoDA(["Col1"])
        twoda.add_row("0", {"Col1": "X"})

        memory = PatcherMemory()
        config = Modifications2DA("")
        config.modifiers.append(AddRow2DA("", "Col1", "1", {"Col1": RowValueConstant("X")}, store_2da={5: RowValueRowIndex()}))
        config.modifiers.append(AddRow2DA("", None, "2", {"Col1": RowValueConstant("Y")}, store_2da={6: RowValueRowIndex()}))
        config.apply(twoda, memory)

        self.assertEqual(2, twoda.get_height())
        self.assertEqual(["X", "Y"], twoda.get_column("Col1"))
        self.assertEqual("0", memory.memory_2da[5])
        self.assertEqual("1", memory.memory_2da[6])
    # endregion

    # region Copy Row
    def test_copy_existing_rowindex(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), None, None, {"Col2": RowValueConstant("X")}))
        config.apply(twoda, memory)

        self.assertEqual(3, twoda.get_height())
        self.assertEqual(["a", "c", "a"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d", "X"], twoda.get_column("Col2"))

    def test_copy_existing_rowlabel(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_LABEL, "1"), None, None, {"Col2": RowValueConstant("X")}))
        config.apply(twoda, memory)

        self.assertEqual(3, twoda.get_height())
        self.assertEqual(["a", "c", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d", "X"], twoda.get_column("Col2"))

    def test_copy_exclusive_notexists(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), "Col1", None, {"Col1": RowValueConstant("c"), "Col2": RowValueConstant("d")}))
        config.apply(twoda, memory)

        self.assertEqual(2, twoda.get_height())
        self.assertEqual("1", twoda.get_label(1))
        self.assertEqual(["a", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d"], twoda.get_column("Col2"))

    def test_copy_exclusive_exists(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), "Col1", None, {"Col1": RowValueConstant("a"), "Col2": RowValueConstant("X")}))
        config.apply(twoda, memory)

        self.assertEqual(1, twoda.get_height())
        self.assertEqual("0", twoda.get_label(0))
        self.assertEqual(["a"], twoda.get_column("Col1"))
        self.assertEqual(["X"], twoda.get_column("Col2"))

    def test_copy_exclusive_none(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), None, None, {"Col1": RowValueConstant("c"), "Col2": RowValueConstant("d")}))
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), "", "r2", {"Col1": RowValueConstant("e"), "Col2": RowValueConstant("f")}))
        config.apply(twoda, memory)

        self.assertEqual(3, twoda.get_height())
        self.assertEqual("1", twoda.get_label(1))
        self.assertEqual("r2", twoda.get_label(2))
        self.assertEqual(["a", "c", "e"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d", "f"], twoda.get_column("Col2"))

    def test_copy_set_newrowlabel(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), None, "r2", {}))
        config.apply(twoda, memory)

        self.assertEqual("r2", twoda.get_label(2))
        self.assertEqual(["a", "c", "a"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d", "b"], twoda.get_column("Col2"))

    def test_copy_assign_high(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "1"})
        twoda.add_row("1", {"Col1": "c", "Col2": "2"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), None, None, {"Col2": RowValueHigh("Col2")}))
        config.apply(twoda, memory)

        self.assertEqual(3, twoda.get_height())
        self.assertEqual(["a", "c", "a"], twoda.get_column("Col1"))
        self.assertEqual(["1", "2", "3"], twoda.get_column("Col2"))

    def test_copy_assign_tlkmemory(self):
        twoda = TwoDA(["Col1", "Col2", "Col3"])
        twoda.add_row("0", {"Col1": "a", "Col2": "1"})
        twoda.add_row("1", {"Col1": "c", "Col2": "2"})

        memory = PatcherMemory()
        memory.memory_str[0] = 5

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), None, None, {"Col2": RowValueTLKMemory(0)}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c", "a"], twoda.get_column("Col1"))
        self.assertEqual(["1", "2", "5"], twoda.get_column("Col2"))

    def test_copy_assign_2damemory(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "1"})
        twoda.add_row("1", {"Col1": "c", "Col2": "2"})

        memory = PatcherMemory()
        memory.memory_2da[0] = "5"

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), None, None, {"Col2": RowValue2DAMemory(0)}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c", "a"], twoda.get_column("Col1"))
        self.assertEqual(["1", "2", "5"], twoda.get_column("Col2"))

    def test_copy_2damemory_rowindex(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(CopyRow2DA("", Target(TargetType.ROW_INDEX, 0), None, None, {}, store_2da={5: RowValueRowIndex()}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c", "a"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d", "b"], twoda.get_column("Col2"))
        self.assertEqual("2", memory.memory_2da[5])
    # endregion

    # region Add Column
    def test_addcolumn_empty(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(AddColumn2DA("", "Col3", "", {}, {}, {}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d"], twoda.get_column("Col2"))
        self.assertEqual(["", ""], twoda.get_column("Col3"))

    def test_addcolumn_default(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(AddColumn2DA("", "Col3", "X", {}, {}, {}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d"], twoda.get_column("Col2"))
        self.assertEqual(["X", "X"], twoda.get_column("Col3"))

    def test_addcolumn_rowindex_constant(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(AddColumn2DA("", "Col3", "", {0: RowValueConstant("X")}, {}, {}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d"], twoda.get_column("Col2"))
        self.assertEqual(["X", ""], twoda.get_column("Col3"))

    def test_addcolumn_rowlabel_2damemory(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()
        memory.memory_2da[5] = "ABC"

        config = Modifications2DA("")
        config.modifiers.append(AddColumn2DA("", "Col3", "", {}, {"1": RowValue2DAMemory(5)}, {}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d"], twoda.get_column("Col2"))
        self.assertEqual(["", "ABC"], twoda.get_column("Col3"))

    def test_addcolumn_rowlabel_tlkmemory(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()
        memory.memory_str[5] = 123

        config = Modifications2DA("")
        config.modifiers.append(AddColumn2DA("", "Col3", "", {}, {"1": RowValueTLKMemory(5)}, {}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d"], twoda.get_column("Col2"))
        self.assertEqual(["", "123"], twoda.get_column("Col3"))

    def test_addcolumn_2damemory_index(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(AddColumn2DA("", "Col3", "", {0: RowValueConstant("X"), 1: RowValueConstant("Y")}, {}, store_2da={0: "I0"}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d"], twoda.get_column("Col2"))
        self.assertEqual(["X", "Y"], twoda.get_column("Col3"))
        self.assertEqual("X", memory.memory_2da[0])

    def test_addcolumn_2damemory_line(self):
        twoda = TwoDA(["Col1", "Col2"])
        twoda.add_row("0", {"Col1": "a", "Col2": "b"})
        twoda.add_row("1", {"Col1": "c", "Col2": "d"})

        memory = PatcherMemory()

        config = Modifications2DA("")
        config.modifiers.append(AddColumn2DA("", "Col3", "", {0: RowValueConstant("X"), 1: RowValueConstant("Y")}, {}, store_2da={0: "L1"}))
        config.apply(twoda, memory)

        self.assertEqual(["a", "c"], twoda.get_column("Col1"))
        self.assertEqual(["b", "d"], twoda.get_column("Col2"))
        self.assertEqual(["X", "Y"], twoda.get_column("Col3"))
        self.assertEqual("Y", memory.memory_2da[0])

    # endregion


class TestManipulateGFF(TestCase):
    def test_modify_field_uint8(self):
        gff = GFF()
        gff.root.set_uint8("Field1", 1)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2))])
        config.apply(gff, memory)

        self.assertEqual(2, gff.root.get_uint8("Field1"))

    def test_modify_field_int8(self):
        gff = GFF()
        gff.root.set_int8("Field1", 1)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2))])
        config.apply(gff, memory)

        self.assertEqual(2, gff.root.get_int8("Field1"))

    def test_modify_field_uint16(self):
        gff = GFF()
        gff.root.set_uint16("Field1", 1)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2))])
        config.apply(gff, memory)

        self.assertEqual(2, gff.root.get_uint16("Field1"))

    def test_modify_field_int16(self):
        gff = GFF()
        gff.root.set_int16("Field1", 1)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2))])
        config.apply(gff, memory)

        self.assertEqual(2, gff.root.get_int16("Field1"))

    def test_modify_field_uint32(self):
        gff = GFF()
        gff.root.set_uint32("Field1", 1)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2))])
        config.apply(gff, memory)

        self.assertEqual(2, gff.root.get_uint32("Field1"))

    def test_modify_field_int32(self):
        gff = GFF()
        gff.root.set_int32("Field1", 1)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2))])
        config.apply(gff, memory)

        self.assertEqual(2, gff.root.get_int32("Field1"))

    def test_modify_field_uint64(self):
        gff = GFF()
        gff.root.set_uint64("Field1", 1)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2))])
        config.apply(gff, memory)

        self.assertEqual(2, gff.root.get_uint64("Field1"))

    def test_modify_field_int64(self):
        gff = GFF()
        gff.root.set_int64("Field1", 1)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2))])
        config.apply(gff, memory)

        self.assertEqual(2, gff.root.get_int64("Field1"))

    def test_modify_field_single(self):
        gff = GFF()
        gff.root.set_single("Field1", 1.234)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2.345))])
        config.apply(gff, memory)

        self.assertEqual(2.345, gff.root.get_single("Field1"))

    def test_modify_field_double(self):
        gff = GFF()
        gff.root.set_double("Field1", 1.234567)

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(2.345678))])
        config.apply(gff, memory)

        self.assertEqual(2.345678, gff.root.get_double("Field1"))

    def test_modify_field_string(self):
        gff = GFF()
        gff.root.set_string("Field1", "abc")

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant("def"))])
        config.apply(gff, memory)

        self.assertEqual("def", gff.root.get_string("Field1"))

    def test_modify_field_locstring(self):
        gff = GFF()
        gff.root.set_locstring("Field1", LocalizedString(0))

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(LocalizedStringDelta(FieldValueConstant(1))))])
        config.apply(gff, memory)

        self.assertEqual(1, gff.root.get_locstring("Field1").stringref)

    def test_modify_field_vector3(self):
        gff = GFF()
        gff.root.set_vector3("Field1", Vector3(0, 1, 2))

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(Vector3(1, 2, 3)))])
        config.apply(gff, memory)

        self.assertEqual(Vector3(1, 2, 3), gff.root.get_vector3("Field1"))

    def test_modify_field_vector4(self):
        gff = GFF()
        gff.root.set_vector4("Field1", Vector4(0, 1, 2, 3))

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("Field1", FieldValueConstant(Vector4(1, 2, 3, 4)))])
        config.apply(gff, memory)

        self.assertEqual(Vector4(1, 2, 3, 4), gff.root.get_vector4("Field1"))

    def test_modify_nested(self):
        gff = GFF()
        gff_list = gff.root.set_list("List", GFFList())
        gff_struct = gff_list.add(0)
        gff_struct.set_string("String", "")

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [ModifyFieldGFF("List\\0\\String", FieldValueConstant("abc"))])
        config.apply(gff, memory)

        self.assertEqual("abc", gff_struct.get_string("String"))

    def test_modify_2damemory(self):
        gff = GFF()
        gff.root.set_string("String", "")
        gff.root.set_uint8("Integer", 0)

        memory = PatcherMemory()
        memory.memory_2da[5] = "123"

        config = ModificationsGFF("", False, [])
        config.modifiers.append(ModifyFieldGFF("String", FieldValue2DAMemory(5)))
        config.modifiers.append(ModifyFieldGFF("Integer", FieldValue2DAMemory(5)))
        config.apply(gff, memory)

        self.assertEqual("123", gff.root.get_string("String"))
        self.assertEqual(123, gff.root.get_uint8("Integer"))

    def test_modify_tlkmemory(self):
        gff = GFF()
        gff.root.set_string("String", "")
        gff.root.set_uint8("Integer", 0)

        memory = PatcherMemory()
        memory.memory_str[5] = 123

        config = ModificationsGFF("", False, [])
        config.modifiers.append(ModifyFieldGFF("String", FieldValueTLKMemory(5)))
        config.modifiers.append(ModifyFieldGFF("Integer", FieldValueTLKMemory(5)))
        config.apply(gff, memory)

        self.assertEqual("123", gff.root.get_string("String"))
        self.assertEqual(123, gff.root.get_uint8("Integer"))

    def test_add_newnested(self):
        gff = GFF()

        memory = PatcherMemory()

        add_field1 = AddFieldGFF("", "List", GFFFieldType.List, FieldValueConstant(GFFList()))

        add_field2 = AddStructToListGFF("", 0)
        add_field1.modifiers.append(add_field2)

        add_field3 = AddFieldGFF("", "SomeInteger", GFFFieldType.UInt8, FieldValueConstant(123))
        add_field2.modifiers.append(add_field3)

        config = ModificationsGFF("", False, [add_field1])
        config.apply(gff, memory)

        self.assertIsNotNone(gff.root.get_list("List"))
        self.assertIsNotNone(gff.root.get_list("List").at(0))
        self.assertIsNotNone(gff.root.get_list("List").at(0).get_uint8("SomeInteger"))

    def test_add_nested(self):
        gff = GFF()
        gff_list = gff.root.set_list("List", GFFList())
        gff_struct = gff_list.add(0)

        memory = PatcherMemory()
        memory.memory_str[5] = 123

        config = ModificationsGFF("", False, [])
        config.modifiers.append(AddFieldGFF("", "String", GFFFieldType.String, FieldValueConstant("abc"), path="List\\0"))
        config.apply(gff, memory)

        self.assertEqual("abc", gff_struct.get_string("String"))

    def test_add_use_2damemory(self):
        gff = GFF()

        memory = PatcherMemory()
        memory.memory_2da[5] = "123"

        config = ModificationsGFF("", False, [])
        config.modifiers.append(AddFieldGFF("", "String", GFFFieldType.String, FieldValue2DAMemory(5)))
        config.modifiers.append(AddFieldGFF("", "Integer", GFFFieldType.UInt8, FieldValue2DAMemory(5)))
        config.apply(gff, memory)

        self.assertEqual("123", gff.root.get_string("String"))
        self.assertEqual(123, gff.root.get_uint8("Integer"))

    def test_add_use_tlkmemory(self):
        gff = GFF()

        memory = PatcherMemory()
        memory.memory_str[5] = 123

        config = ModificationsGFF("", False, [])
        config.modifiers.append(AddFieldGFF("", "String", GFFFieldType.String, FieldValueTLKMemory(5)))
        config.modifiers.append(AddFieldGFF("", "Integer", GFFFieldType.UInt8, FieldValueTLKMemory(5)))
        config.apply(gff, memory)

        self.assertEqual("123", gff.root.get_string("String"))
        self.assertEqual(123, gff.root.get_uint8("Integer"))

    def test_add_field_locstring(self):
        gff = GFF()
        gff.root.set_locstring("Field1", LocalizedString(0))

        memory = PatcherMemory()
        memory.memory_2da[5] = "123"

        config = ModificationsGFF("", False, [AddFieldGFF("", "Field1", GFFFieldType.LocalizedString, FieldValueConstant(LocalizedStringDelta(FieldValue2DAMemory(5))))])
        config.apply(gff, memory)

        self.assertEqual(123, gff.root.get_locstring("Field1").stringref)

    def test_addlist_listindex(self):
        gff = GFF()
        gff_list = gff.root.set_list("List", GFFList())

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [])
        config.modifiers.append(AddStructToListGFF("", None, "List"))
        config.modifiers.append(AddStructToListGFF("", None, "List"))
        config.modifiers.append(AddStructToListGFF("", None, "List"))
        config.apply(gff, memory)

        self.assertEqual(0, gff_list.at(0).struct_id)
        self.assertEqual(1, gff_list.at(1).struct_id)
        self.assertEqual(2, gff_list.at(2).struct_id)

    def test_addlist_store_2damemory(self):
        gff = GFF()
        gff.root.set_list("List", GFFList())

        memory = PatcherMemory()

        config = ModificationsGFF("", False, [])
        config.modifiers.append(AddStructToListGFF("", 0, path="List"))
        config.modifiers.append(AddStructToListGFF("", 0, path="List", index_to_token=12))
        config.apply(gff, memory)

        self.assertEqual("1", memory.memory_2da[12])


class TestManipulateSSF(TestCase):
    def test_assign_int(self):
        ssf = SSF()

        memory = PatcherMemory()

        config = ModificationsSSF("", False, [])
        config.modifiers.append(ModifySSF(SSFSound.BATTLE_CRY_1, NoTokenUsage(5)))
        config.apply(ssf, memory)

        self.assertEqual(5, ssf.get(SSFSound.BATTLE_CRY_1))

    def test_assign_2datoken(self):
        ssf = SSF()

        memory = PatcherMemory()
        memory.memory_2da[5] = "123"

        config = ModificationsSSF("", False, [])
        config.modifiers.append(ModifySSF(SSFSound.BATTLE_CRY_2, TokenUsage2DA(5)))
        config.apply(ssf, memory)

        self.assertEqual(123, ssf.get(SSFSound.BATTLE_CRY_2))

    def test_assign_tlktoken(self):
        ssf = SSF()

        memory = PatcherMemory()
        memory.memory_str[5] = 321

        config = ModificationsSSF("", False, [])
        config.modifiers.append(ModifySSF(SSFSound.BATTLE_CRY_3, TokenUsageTLK(5)))
        config.apply(ssf, memory)

        self.assertEqual(321, ssf.get(SSFSound.BATTLE_CRY_3))




import os.path
from abc import ABC, abstractmethod
from typing import List

from pykotor.common.stream import BinaryReader, BinaryWriter

from pykotor.extract.file import ResourceIdentifier

from pykotor.extract.capsule import Capsule

from pykotor.tools.misc import is_capsule_file
from pykotor.tslpatcher.logger import PatchLogger


class InstallFile:
    def __init__(self, filename: str, replace_existing: bool):
        self.filename: str = filename
        self.replace_existing: bool = replace_existing

    def _identifier(self) -> ResourceIdentifier:
        return ResourceIdentifier.from_path(self.filename)

    def apply_encapsulated(self, log: PatchLogger, source_folder: str, destination: Capsule):
        resname, restype = self._identifier()

        if self.replace_existing or destination.resource(resname, restype) is None:
            if self.replace_existing and destination.resource(resname, restype) is not None:
                log.add_note("Replacing file {} in the {} archive...".format(self.filename, destination.path()))
            else:
                log.add_note("Adding file {} in the {} archive...".format(self.filename, destination.path()))

            data = BinaryReader.load_file("{}/{}".format(source_folder, self.filename))
            destination.add(resname, restype, data)

    def apply_file(self, log: PatchLogger, source_folder: str, destination: str):
        data = BinaryReader.load_file("{}/{}".format(source_folder, self.filename))
        save_file_to = "{}/{}".format(destination, self.filename)

        if self.replace_existing or not os.path.exists(save_file_to):
            if not os.path.exists(destination):
                log.add_note("Folder {} did not exist, creating it...".format(destination))
                os.makedirs(destination)

            if self.replace_existing and not os.path.exists(save_file_to):
                log.add_note("Replacing file {} to the {} folder...".format(self.filename, destination))
            else:
                log.add_note("Copying file {} to the {} folder...".format(self.filename, destination))

            BinaryWriter.dump(save_file_to, data)
        elif not self.replace_existing and os.path.exists(save_file_to):
            log.add_warning("A file named {} already exists in the {} folder. Skipping file...".format(self.filename, destination))


class InstallFolder:
    def __init__(self, foldername: str, files: List[InstallFile] = None):
        self.foldername: str = foldername
        self.files: List[InstallFile] = [] if files is None else files

    def apply(self, log: PatchLogger, source_path: str, destination_path: str):
        target = "{}/{}".format(destination_path, self.foldername)

        if is_capsule_file(self.foldername):
            destination = Capsule(target, create_nonexisting=True)
            [file.apply_encapsulated(log, source_path, destination) for file in self.files]
        else:
            [file.apply_file(log, source_path, target) for file in self.files]



import platform
from unittest import TestCase

from pykotor.resource.formats.ssf import SSF, SSFSound, SSFBinaryReader, detect_ssf, SSFXMLReader, write_ssf, read_ssf
from pykotor.resource.type import ResourceType


BINARY_TEST_FILE = "../../files/test.ssf"
XML_TEST_FILE = "../../files/test.ssf.xml"
DOES_NOT_EXIST_FILE = "./thisfiledoesnotexist"
CORRUPT_BINARY_TEST_FILE = "../../files/test_corrupted.ssf"
CORRUPT_XML_TEST_FILE = "../../files/test_corrupted.ssf.xml"


class TestSSF(TestCase):
    def test_binary_io(self):
        self.assertEqual(detect_ssf(BINARY_TEST_FILE), ResourceType.SSF)

        ssf = SSFBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(ssf)

        data = bytearray()
        write_ssf(ssf, data, ResourceType.SSF)
        ssf = SSFBinaryReader(data).load()
        self.validate_io(ssf)

    def test_xml_io(self):
        self.assertEqual(detect_ssf(XML_TEST_FILE), ResourceType.SSF_XML)

        ssf = SSFXMLReader(XML_TEST_FILE).load()
        self.validate_io(ssf)

        data = bytearray()
        write_ssf(ssf, data, ResourceType.SSF_XML)
        ssf = SSFXMLReader(data).load()
        self.validate_io(ssf)

    def validate_io(self, ssf: SSF):
        self.assertEqual(ssf.get(SSFSound.BATTLE_CRY_1), 123075)
        self.assertEqual(ssf.get(SSFSound.BATTLE_CRY_2), 123074)
        self.assertEqual(ssf.get(SSFSound.BATTLE_CRY_3), 123073)
        self.assertEqual(ssf.get(SSFSound.BATTLE_CRY_4), 123072)
        self.assertEqual(ssf.get(SSFSound.BATTLE_CRY_5), 123071)
        self.assertEqual(ssf.get(SSFSound.BATTLE_CRY_6), 123070)
        self.assertEqual(ssf.get(SSFSound.SELECT_1), 123069)
        self.assertEqual(ssf.get(SSFSound.SELECT_2), 123068)
        self.assertEqual(ssf.get(SSFSound.SELECT_3), 123067)
        self.assertEqual(ssf.get(SSFSound.ATTACK_GRUNT_1), 123066)
        self.assertEqual(ssf.get(SSFSound.ATTACK_GRUNT_2), 123065)
        self.assertEqual(ssf.get(SSFSound.ATTACK_GRUNT_3), 123064)
        self.assertEqual(ssf.get(SSFSound.PAIN_GRUNT_1), 123063)
        self.assertEqual(ssf.get(SSFSound.PAIN_GRUNT_2), 123062)
        self.assertEqual(ssf.get(SSFSound.LOW_HEALTH), 123061)
        self.assertEqual(ssf.get(SSFSound.DEAD), 123060)
        self.assertEqual(ssf.get(SSFSound.CRITICAL_HIT), 123059)
        self.assertEqual(ssf.get(SSFSound.TARGET_IMMUNE), 123058)
        self.assertEqual(ssf.get(SSFSound.LAY_MINE), 123057)
        self.assertEqual(ssf.get(SSFSound.DISARM_MINE), 123056)
        self.assertEqual(ssf.get(SSFSound.BEGIN_STEALTH), 123055)
        self.assertEqual(ssf.get(SSFSound.BEGIN_SEARCH), 123054)
        self.assertEqual(ssf.get(SSFSound.BEGIN_UNLOCK), 123053)
        self.assertEqual(ssf.get(SSFSound.UNLOCK_FAILED), 123052)
        self.assertEqual(ssf.get(SSFSound.UNLOCK_SUCCESS), 123051)
        self.assertEqual(ssf.get(SSFSound.SEPARATED_FROM_PARTY), 123050)
        self.assertEqual(ssf.get(SSFSound.REJOINED_PARTY), 123049)
        self.assertEqual(ssf.get(SSFSound.POISONED), 123048)

    def test_read_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, read_ssf, ".")
        else:
            self.assertRaises(IsADirectoryError, read_ssf, ".")
        self.assertRaises(FileNotFoundError, read_ssf, DOES_NOT_EXIST_FILE)
        self.assertRaises(ValueError, read_ssf, CORRUPT_BINARY_TEST_FILE)
        self.assertRaises(ValueError, read_ssf, CORRUPT_XML_TEST_FILE)

    def test_write_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, write_ssf, SSF(), ".", ResourceType.SSF)
        else:
            self.assertRaises(IsADirectoryError, write_ssf, SSF(), ".", ResourceType.SSF)
        self.assertRaises(ValueError, write_ssf, SSF(), ".", ResourceType.INVALID)



from pykotor.resource.formats.rim.rim_data import RIM, RIMResource
from pykotor.resource.formats.rim.io_rim import RIMBinaryReader, RIMBinaryWriter
from pykotor.resource.formats.rim.rim_auto import read_rim, write_rim



from pykotor.resource.formats.tpc.tpc_data import TPC, TPCTextureFormat
from pykotor.resource.formats.tpc.io_tpc import TPCBinaryReader, TPCBinaryWriter
from pykotor.resource.formats.tpc.io_tga import TPCTGAWriter, TPCTGAReader
from pykotor.resource.formats.tpc.io_bmp import TPCBMPWriter
from pykotor.resource.formats.tpc.tpc_auto import read_tpc, write_tpc, detect_tpc



import struct
from enum import IntEnum
from typing import Optional

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.tpc import TPC, TPCTextureFormat
from pykotor.resource.type import ResourceWriter, TARGET_TYPES, ResourceReader, SOURCE_TYPES, autoclose


class _DataTypes(IntEnum):
    NO_IMAGE_DATA = 0
    UNCOMPRESSED_COLOR_MAPPED = 1
    UNCOMPRESSED_RGB = 2
    UNCOMPRESSED_BLACK_WHITE = 3
    RLE_COLOR_MAPPED = 9
    RLE_RGB = 10
    COMPRESSED_BLACK_WHITE = 11
    COMPRESSED_COLOR_MAPPED_A = 32
    COMPRESSED_COLOR_MAPPED_B = 33


class TPCTGAReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._tpc: Optional[TPC] = None

    #@autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> TPC:
        self._tpc = TPC()

        id_length = self._reader.read_uint8()
        colormap_type = self._reader.read_uint8()
        datatype_code = self._reader.read_uint8()
        colormap_origin = self._reader.read_uint16()
        colormap_length = self._reader.read_uint16()
        colormap_depth = self._reader.read_uint8()
        x_origin = self._reader.read_uint16()
        y_origin = self._reader.read_uint16()
        width = self._reader.read_uint16()
        height = self._reader.read_uint16()
        bits_per_pixel = self._reader.read_uint8()
        image_descriptor = self._reader.read_uint8()
        self._reader.skip(id_length)

        y_flipped = bool(image_descriptor & 0b00100000)
        interleaving_id = (image_descriptor & 0b11000000) >> 6

        if interleaving_id:
            ValueError("The image data must not be interleaved.")

        if datatype_code == _DataTypes.UNCOMPRESSED_RGB:
            self._reader.skip(colormap_length * colormap_depth // 8)
            data = bytearray()

            if bits_per_pixel != 24 and bits_per_pixel != 32:
                ValueError("The image must store 24 or 32 bits per pixel.")

            pixel_rows = []
            for y in range(height):
                pixel_rows.append(bytearray())
                for x in range(width):
                    b, g, r = self._reader.read_uint8(), self._reader.read_uint8(), self._reader.read_uint8()
                    if bits_per_pixel == 32:
                        pixel_rows[y].extend([r, g, b, self._reader.read_uint8()])
                    else:
                        pixel_rows[y].extend([r, g, b])

            if y_flipped:
                [data.extend(pixels) for pixels in reversed(pixel_rows)]
            else:
                [data.extend(pixels) for pixels in pixel_rows]
        elif datatype_code == _DataTypes.RLE_RGB:
            data = bytearray()
            pixels = []
            n = 0

            while self._reader.remaining():
                packet = self._reader.read_uint8()

                raw = (packet >> 7) == 0
                count = (packet & 0b01111111) + 1
                n += count

                if raw:
                    for i in range(count):
                        b, g, r = self._reader.read_uint8(), self._reader.read_uint8(), self._reader.read_uint8()
                        if bits_per_pixel == 32:
                            pixels.extend([r, g, b, self._reader.read_uint8()])
                        else:
                            pixels.extend([r, g, b, 255])
                else:
                    b, g, r = self._reader.read_uint8(), self._reader.read_uint8(), self._reader.read_uint8()
                    pixel = [r, g, b]
                    if bits_per_pixel == 32:
                        pixel.append(self._reader.read_uint8())
                    else:
                        pixel.append(255)
                    for i in range(count):
                        pixels.extend(pixel)

                if n == width*height:
                    break
            data.extend(pixels)
        else:
            raise ValueError("The image data must be RGB and not color-mapped.")

        self._tpc.set(width, height, [bytes(data)], TPCTextureFormat.RGBA)

        return self._tpc


class TPCTGAWriter(ResourceWriter):
    def __init__(
            self,
            tpc: TPC,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._tpc = tpc

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        width, height = self._tpc.dimensions()

        self._writer.write_uint8(0)
        self._writer.write_uint8(0)
        self._writer.write_uint8(2)
        self._writer.write_bytes(bytes(5))
        self._writer.write_uint16(0)
        self._writer.write_uint16(0)
        self._writer.write_uint16(width)
        self._writer.write_uint16(height)

        if self._tpc.format() in [TPCTextureFormat.RGB or TPCTextureFormat.DXT1]:
            self._writer.write_uint8(32)
            self._writer.write_uint8(0)
            data = self._tpc.convert(TPCTextureFormat.RGB, 0).data
            pixel_reader = BinaryReader.from_bytes(data)
            for i in range(len(data) // 3):
                r = pixel_reader.read_uint8()
                g = pixel_reader.read_uint8()
                b = pixel_reader.read_uint8()
                self._writer.write_bytes(struct.pack('BBBB', b, g, r, 255))
        else:
            self._writer.write_uint8(32)
            self._writer.write_uint8(0)
            width, height, data = self._tpc.convert(TPCTextureFormat.RGBA, 0)
            pixel_reader = BinaryReader.from_bytes(data)
            for i in range(len(data) // 4):
                r = pixel_reader.read_uint8()
                g = pixel_reader.read_uint8()
                b = pixel_reader.read_uint8()
                a = pixel_reader.read_uint8()
                self._writer.write_bytes(struct.pack('BBBB', b, g, r, a))



from pykotor.resource.formats.twoda import TwoDA


class Diff2DA:
    def __init__(self, old: TwoDA, new: TwoDA):
        self.old: TwoDA = old
        self.new: TwoDA = new

    def is_same(self) -> bool:
        old_columns = set(self.old.get_headers())
        new_columns = set(self.new.get_headers())

        if old_columns != new_columns:
            print("Columns match")
            return False

        for old_row in self.old:
            index = self.old.row_index(old_row)
            new_row = self.new.get_row(index)

            for header in old_columns:
                old_value = old_row.get_string(header)
                new_value = new_row.get_string(header)
                if old_value != new_value:
                    print("Cell mismatch:", index, header)
                    return False

        return True



from __future__ import annotations

from typing import List

from pykotor.common.language import LocalizedString
from pykotor.common.misc import ResRef, Game, InventoryItem
from pykotor.resource.formats.gff import GFF, GFFList, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, TARGET_TYPES, SOURCE_TYPES


class UTP:
    """
    Stores placeable data.

    Attributes:
        tag: "Tag" field.
        name: "LocName" field.
        resref: "TemplateResRef" field.
        auto_remove_key: "AutoRemoveKey" field.
        conversation: "Conversation" field.
        faction_id: "Faction" field.
        plot: "Plot" field.
        min1_hp: "Min1HP" field.
        key_required: "KeyRequired" field.
        lockable: "Lockable" field.
        locked: "Locked" field.
        unlock_dc: "OpenLockDC" field.
        key_name: "KeyName" field.
        animation_state: "AnimationState" field.
        appearance_id: "Appearance" field.
        maximum_hp: "HP" field.
        current_hp: "CurrentHP" field.
        hardness: "Hardness" field.
        fortitude: "Fort" field.
        on_closed: "OnClosed" field.
        on_damaged: "OnDamaged" field.
        on_death: "OnDeath" field.
        on_heartbeat: "OnHeartbeat" field.
        on_lock: "OnLock" field.
        on_melee_attack: "OnMeleeAttacked" field.
        on_open: "OnOpen" field.
        on_force_power: "OnSpellCastAt" field.
        on_unlock: "OnUnlock" field.
        on_user_defined: "OnUserDefined" field.
        has_inventory: "HasInventory" field.
        party_interact: "PartyInteract" field.
        static: "Static" field.
        useable: "Useable" field.
        on_end_dialog: "OnEndDialogue" field.
        on_inventory: "OnInvDisturbed" field.
        on_used: "OnUsed" field.
        comment: "Comment" field.

        not_blastable: "NotBlastable" field. KotOR 2 Only.
        unlock_diff: "OpenLockDiff" field. KotOR 2 Only.
        unlock_diff_mod: "OpenLockDiffMod" field. KotOR 2 Only.
        on_open_failed: "OnFailToOpen" field. KotOR 2 Only.
        lock_dc: "CloseLockDC" field. KotOR 2 Only.

        palette_id: "PaletteID" field. Used in toolset only.

        description: "Description" field. Not used by the game engine.
        interruptable: "Interruptable" field. Not used by the game engine.
        portrait_id: "PortraitId" field. Not used by the game engine.
        trap_detectable: "TrapDetectable" field. Not used by the game engine.
        trap_detect_dc: "TrapDetectDC" field. Not used by the game engine.
        trap_disarmable: "TrapDisarmable" field. Not used by the game engine.
        trap_disarm_dc: "DisarmDC" field. Not used by the game engine.
        trap_flag: "TrapFlag" field. Not used by the game engine.
        trap_one_shot: "TrapOneShot" field. Not used by the game engine.
        trap_type: "TrapType" field. Not used by the game engine.
        will: "Will" field. Not used by the game engine.
        on_disarm: "OnDisarm" field. Not used by the game engine.
        on_trap_triggered: "OnTrapTriggered" field. Not used by the game engine.
        bodybag_id: "BodyBag" field. Not used by the game engine.
        type_id: "Type" field. Not used by the game engine.
        lock_dc: "CloseLockDC" field. Not used by the game engine.
    """

    BINARY_TYPE = ResourceType.UTP

    def __init__(
            self
    ):
        self.resref: ResRef = ResRef.from_blank()
        self.conversation: ResRef = ResRef.from_blank()
        self.tag: str = ""
        self.comment: str = ""

        self.name: LocalizedString = LocalizedString.from_invalid()

        self.faction_id: int = 0
        self.appearance_id: int = 0
        self.animation_state: int = 0

        self.auto_remove_key: bool = False
        self.key_name: str = ""
        self.key_required: bool = False
        self.locked: bool = False
        self.lockable: bool = False

        self.unlock_dc: int = 0
        self.unlock_diff: int = 0  # KotOR 2 Only
        self.unlock_diff_mod: int = 0  # KotOR 2 Only

        self.current_hp: int = 0
        self.maximum_hp: int = 0
        self.fortitude: int = 0
        self.hardness: int = 0

        self.min1_hp: bool = False
        self.not_blastable: bool = False
        self.plot: bool = False
        self.static: bool = False
        self.useable: bool = False
        self.party_interact: bool = False

        self.on_closed: ResRef = ResRef.from_blank()
        self.on_damaged: ResRef = ResRef.from_blank()
        self.on_death: ResRef = ResRef.from_blank()
        self.on_end_dialog: ResRef = ResRef.from_blank()
        self.on_open_failed: ResRef = ResRef.from_blank()
        self.on_heartbeat: ResRef = ResRef.from_blank()
        self.on_inventory: ResRef = ResRef.from_blank()
        self.on_melee_attack: ResRef = ResRef.from_blank()
        self.on_force_power: ResRef = ResRef.from_blank()
        self.on_open: ResRef = ResRef.from_blank()
        self.on_lock: ResRef = ResRef.from_blank()
        self.on_unlock: ResRef = ResRef.from_blank()
        self.on_used: ResRef = ResRef.from_blank()
        self.on_user_defined: ResRef = ResRef.from_blank()

        self.has_inventory: bool = False
        self.inventory: List[InventoryItem] = []

        # Deprecated:
        self.description: LocalizedString = LocalizedString.from_invalid()
        self.interruptable: bool = False
        self.portrait_id: int = 0
        self.trap_detectable: bool = False
        self.trap_detect_dc: int = 0
        self.trap_disarmable: bool = False
        self.trap_disarm_dc: int = 0
        self.trap_flag: int = 0
        self.trap_one_shot: bool = False
        self.trap_type: int = 0
        self.will: int = 0
        self.reflex: int = 0
        self.on_disarm: ResRef = ResRef.from_blank()
        self.on_trap_triggered: ResRef = ResRef.from_blank()
        self.bodybag_id: int = 0
        self.type_id: int = 0
        self.palette_id: int = 0
        self.lock_dc: int = 0


def construct_utp(
        gff: GFF
) -> UTP:
    utp = UTP()

    root = gff.root
    utp.tag = root.acquire("Tag", "")
    utp.name = root.acquire("LocName", LocalizedString.from_invalid())
    utp.resref = root.acquire("TemplateResRef", ResRef.from_blank())
    utp.auto_remove_key = bool(root.acquire("AutoRemoveKey", 0))
    utp.lock_dc = root.acquire("CloseLockDC", 0)
    utp.conversation = root.acquire("Conversation", ResRef.from_blank())
    utp.faction_id = root.acquire("Faction", 0)
    utp.plot = bool(root.acquire("Plot", 0))
    utp.not_blastable = bool(root.acquire("NotBlastable", 0))
    utp.min1_hp = bool(root.acquire("Min1HP", 0))
    utp.key_required = bool(root.acquire("KeyRequired", 0))
    utp.lockable = bool(root.acquire("Lockable", 0))
    utp.locked = bool(root.acquire("Locked", 0))
    utp.unlock_dc = root.acquire("OpenLockDC", 0)
    utp.key_name = root.acquire("KeyName", "")
    utp.animation_state = root.acquire("AnimationState", 0)
    utp.appearance_id = root.acquire("Appearance", 0)
    utp.maximum_hp = root.acquire("HP", 0)
    utp.current_hp = root.acquire("CurrentHP", 0)
    utp.hardness = root.acquire("Hardness", 0)
    utp.fortitude = root.acquire("Fort", 0)
    utp.on_closed = root.acquire("OnClosed", ResRef.from_blank())
    utp.on_damaged = root.acquire("OnDamaged", ResRef.from_blank())
    utp.on_death = root.acquire("OnDeath", ResRef.from_blank())
    utp.on_heartbeat = root.acquire("OnHeartbeat", ResRef.from_blank())
    utp.on_lock = root.acquire("OnLock", ResRef.from_blank())
    utp.on_melee_attack = root.acquire("OnMeleeAttacked", ResRef.from_blank())
    utp.on_open = root.acquire("OnOpen", ResRef.from_blank())
    utp.on_force_power = root.acquire("OnSpellCastAt", ResRef.from_blank())
    utp.on_unlock = root.acquire("OnUnlock", ResRef.from_blank())
    utp.on_user_defined = root.acquire("OnUserDefined", ResRef.from_blank())
    utp.has_inventory = bool(root.acquire("HasInventory", 0))
    utp.party_interact = bool(root.acquire("PartyInteract", 0))
    utp.static = bool(root.acquire("Static", 0))
    utp.useable = bool(root.acquire("Useable", 0))
    utp.on_end_dialog = root.acquire("OnEndDialogue", ResRef.from_blank())
    utp.on_inventory = root.acquire("OnInvDisturbed", ResRef.from_blank())
    utp.on_used = root.acquire("OnUsed", ResRef.from_blank())
    utp.on_open_failed = root.acquire("OnFailToOpen", ResRef.from_blank())
    utp.comment = root.acquire("Comment", "")
    utp.unlock_diff = root.acquire("OpenLockDiff", 0)
    utp.unlock_diff_mod = root.acquire("OpenLockDiffMod", 0)

    item_list: GFFList = root.acquire("ItemList", GFFList())
    for item_struct in item_list:
        resref = item_struct.acquire("InventoryRes", ResRef.from_blank())
        droppable = bool(item_struct.acquire("Droppable", 0))
        utp.inventory.append(InventoryItem(resref, droppable))

    utp.description = root.acquire("Description", LocalizedString.from_invalid())
    utp.interruptable = root.acquire("Interruptable", 0)
    utp.portrait_id = root.acquire("PortraitId", 0)
    utp.trap_detectable = root.acquire("TrapDetectable", 0)
    utp.trap_detect_dc = root.acquire("TrapDetectDC", 0)
    utp.trap_disarmable = root.acquire("TrapDisarmable", 0)
    utp.trap_disarm_dc = root.acquire("DisarmDC", 0)
    utp.trap_flag = root.acquire("TrapFlag", 0)
    utp.trap_one_shot = root.acquire("TrapOneShot", 0)
    utp.trap_type = root.acquire("TrapType", 0)
    utp.will = root.acquire("Will", 0)
    utp.reflex = root.acquire("Ref", 0)
    utp.on_disarm = root.acquire("OnDisarm", ResRef.from_blank())
    utp.on_trap_triggered = root.acquire("OnTrapTriggered", ResRef.from_blank())
    utp.bodybag_id = root.acquire("BodyBag", 0)
    utp.type_id = root.acquire("Type", 0)
    utp.palette_id = root.acquire("PaletteID", 0)

    return utp


def dismantle_utp(
        utp: UTP,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.UTP)

    root = gff.root
    root.set_string("Tag", utp.tag)
    root.set_locstring("LocName", utp.name)
    root.set_resref("TemplateResRef", utp.resref)
    root.set_uint8("AutoRemoveKey", utp.auto_remove_key)
    root.set_resref("Conversation", utp.conversation)
    root.set_uint32("Faction", utp.faction_id)
    root.set_uint8("Plot", utp.plot)
    root.set_uint8("Min1HP", utp.min1_hp)
    root.set_uint8("KeyRequired", utp.key_required)
    root.set_uint8("Lockable", utp.lockable)
    root.set_uint8("Locked", utp.locked)
    root.set_uint8("OpenLockDC", utp.unlock_dc)
    root.set_string("KeyName", utp.key_name)
    root.set_uint8("AnimationState", utp.animation_state)
    root.set_uint32("Appearance", utp.appearance_id)
    root.set_int16("HP", utp.maximum_hp)
    root.set_int16("CurrentHP", utp.current_hp)
    root.set_uint8("Hardness", utp.hardness)
    root.set_uint8("Fort", utp.fortitude)
    root.set_resref("OnClosed", utp.on_closed)
    root.set_resref("OnDamaged", utp.on_damaged)
    root.set_resref("OnDeath", utp.on_death)
    root.set_resref("OnHeartbeat", utp.on_heartbeat)
    root.set_resref("OnLock", utp.on_lock)
    root.set_resref("OnMeleeAttacked", utp.on_melee_attack)
    root.set_resref("OnOpen", utp.on_open)
    root.set_resref("OnSpellCastAt", utp.on_force_power)
    root.set_resref("OnUnlock", utp.on_unlock)
    root.set_resref("OnUserDefined", utp.on_user_defined)
    root.set_uint8("HasInventory", utp.has_inventory)
    root.set_uint8("PartyInteract", utp.party_interact)
    root.set_uint8("Static", utp.static)
    root.set_uint8("Useable", utp.useable)
    root.set_resref("OnEndDialogue", utp.on_end_dialog)
    root.set_resref("OnInvDisturbed", utp.on_inventory)
    root.set_resref("OnUsed", utp.on_used)
    root.set_string("Comment", utp.comment)

    item_list = root.set_list("ItemList", GFFList())
    for i, item in enumerate(utp.inventory):
        item_struct = item_list.add(i)
        item_struct.set_resref("InventoryRes", item.resref)
        item_struct.set_uint16("Repos_PosX", i)
        item_struct.set_uint16("Repos_Posy", 0)
        if item.droppable:
            item_struct.set_uint8("Droppable", True)

    root.set_uint8("PaletteID", utp.palette_id)

    if game == Game.K2:
        root.set_uint8("NotBlastable", utp.not_blastable)
        root.set_uint8("OpenLockDiff", utp.unlock_diff)
        root.set_uint8("OpenLockDiffMod", utp.unlock_diff_mod)
        root.set_resref("OnFailToOpen", utp.on_open_failed)
        root.set_uint8("CloseLockDC", utp.lock_dc)

    if use_deprecated:
        root.set_locstring("Description", utp.description)
        root.set_uint8("Interruptable", utp.interruptable)
        root.set_uint16("PortraitId", utp.portrait_id)
        root.set_uint8("TrapDetectable", utp.trap_detectable)
        root.set_uint8("TrapDetectDC", utp.trap_detect_dc)
        root.set_uint8("TrapDisarmable", utp.trap_disarmable)
        root.set_uint8("DisarmDC", utp.trap_disarm_dc)
        root.set_uint8("TrapFlag", utp.trap_flag)
        root.set_uint8("TrapOneShot", utp.trap_one_shot)
        root.set_uint8("TrapType", utp.trap_type)
        root.set_uint8("Will", utp.will)
        root.set_uint8("Ref", utp.reflex)
        root.set_resref("OnDisarm", utp.on_disarm)
        root.set_resref("OnTrapTriggered", utp.on_trap_triggered)
        root.set_uint8("BodyBag", utp.bodybag_id)
        root.set_uint8("Type", utp.type_id)

    return gff


def read_utp(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> UTP:
    gff = read_gff(source, offset, size)
    utp = construct_utp(gff)
    return utp


def write_utp(
        utp: UTP,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_utp(utp, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_utp(
        utp: UTP,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_utp(utp, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from pykotor.resource.formats.ltr.ltr_data import LTR, LTRBlock
from pykotor.resource.formats.ltr.ltr_auto import read_ltr, write_ltr
from pykotor.resource.formats.ltr.io_ltr import LTRBinaryReader, LTRBinaryWriter



from __future__ import annotations

import io
from contextlib import suppress
from typing import Optional
from xml.etree import ElementTree

from pykotor.resource.formats.ssf.ssf_data import SSF, SSFSound
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter, autoclose


class SSFXMLReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._ssf: Optional[SSF] = None

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> SSF:
        self._ssf = SSF()

        data = self._reader.read_bytes(self._reader.size()).decode()
        xml_root = ElementTree.parse(io.StringIO(data)).getroot()

        for child in xml_root:
            with suppress(ValueError):
                sound = SSFSound(int(child.attrib['id']))
                stringref = int(child.attrib['strref'])
                self._ssf.set(sound, stringref)

        return self._ssf


class SSFXMLWriter(ResourceWriter):
    def __init__(
            self,
            ssf: SSF,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self.xml_root: ElementTree.Element = ElementTree.Element("xml")
        self.ssf: SSF = ssf

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        for sound_name, sound in SSFSound.__members__.items():
            ElementTree.SubElement(self.xml_root, "sound", {
                "id": str(sound.value),
                "label": sound_name,
                "strref": str(self.ssf.get(sound))
            })

        ElementTree.indent(self.xml_root)
        self._writer.write_bytes(ElementTree.tostring(self.xml_root))
